/*************************************************************************** 
 *
 *                       uhfa.c Version 1.6
 *
 *            Control Panel for DSP-10 2-meter Transciever
 *
 *
 * NOTICE: Copyright (C) 1996, 1997, 1998, 1999 by Robert S. Larkin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The program author can be contacted by email at boblark@proaxis.com.
 * Alternatively contact the author at
 *     Robert S. (Bob) Larkin
 *     2982 N. W. Acacia Place
 *     Corvallis, OR  97330  USA
 *
 * For more details on the background or use of the program, pay a visit to
 * http://www.proaxis.com/~boblark/dsp10.htm  and,
 * http://www.ao.com/~gnome/
 *
 * Compatable with EZKIT-Lite program DSP3 version 0X13 and later.
 *
 * Bob Larkin  W7PUA   Initial write 17 May 96 
 * 29 Oct 96  Added audio FIR, warm boot.  RSL
 * 16 Nov 96  Added audio input and 8/16 bit data option. RSL
 * 18 Nov 96  Added normalization and recording to disk. RSL
 * 10 Dec 96  Reorganized keyboard to allow CW. RSL
 * 22 Dec 96  Added GIF recording of screen image.  RSL
 *  2 Jan 97  Added transverters and configuration file.  RSL
 * 20 Jan 97  Added CW transmit functions.  RSL
 * 27 June 97 Revised to updated hardware: RF gains, T/R sequencing RSL
 * 28 Sept 97 Added control for DAC for carrier cancellation
 * 11 Mar 98  Revised to crystal filter hdw. Removed i-q corrections. RSL
 * Ver1.3:
 * 21 Sept 98 Added experimental versions of LHL and PUA43 Code estimators
 * 25 Nov  98 Revised to a 5 kHz ref freq for 1st LO (v1.31)
 * Ver1.40
 * 17 Jan  99 Full review.  Fixed error 112 problems. Ver 1.4 RSL
 * 17 Apr  99 Update of audio processor unique elements.
 * 29 Apr  99 Corrections & added binaural. Ver 1.41
 *  2 Aug  99 Startup detail, Fd corrected.  v1.5  RSL
 * 12 Sept 99 Put under GNU GPL, corrected startup error.
 *            Removed GIF for current distribution.  V1.6, RSL
 *
 * The ASYNC routines are used to provide buffered serial communications
 * with the DSP.  These are freeware from
 *	David Kessner
 *	david@kessner.denver.co.us
 *	1135 Fairfax
 *	Denver CO  80220
 * Many thanks for use of these routines.
 *
 * Likewise, the GIF routines are greatly appreciated public domain
 * program routines from
 *	Sverre H. Huseby
 *	Bjoelsengt. 17
 *	N-0468 Oslo
 *	Norway
 * who informs us that the following message should follow this program,
 * or any other program borrowing the GIF routines, wherever they may go:
 *
 *	The Graphics Interchange Format(c) is the Copyright property of 
 *	CompuServe Incorporated. GIF(tm) is a Service Mark property of 
 *	CompuServe Incorporated. 
 * Unfortunately, the legal status of incorporating the GIF routines is
 * unclear.  They are not available until this is resolved.
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include <graphics.h>
#include <string.h>
#include <dos.h>
#include <bios.h>
#include <time.h>
#include <alloc.h>
#include <signal.h>
#include "uhfa.h"
#include "uhf_kbd.h"
#include "u_code.h"
#include "async.h"
#include "gifsave.h"
#include "moonsun.h"

/* LOG_OFFSET is sent to dsp to offset log data before
   conversion to 8 bits. Every 4096 is 10 dB. */
#define LOG_OFFSET 12288
#define LOG_OFFSET_D 12288.0

void update_data(void);
void prep_display(void);
void print_stars(void);
void draw_grid(void);
void sa_graph(void);
void mark_frequency(int);
int  do_cmd(void);
void t_r_toggle(void);
void new_freq(void);
void set_afgain(void);
void set_rfgain(void);
void set_mike_gain(void);
void set_fm_squelch(void);
void set_cwoffset(void);
void set_xmit_power(void);
void set_contrast(void);
void set_brightness(void);
void set_window(void);
void set_affilt_sw(void);
void set_affilt_select(void);
void set_delay_right(int);
void filter_design(void);
void select_denoise(void);
void set_denoise(void);
void annotate_af_filt(void);
void set_dot(void);
void set_display_width(void);
void set_disp_offset(int);
void set_db_div(int);
void set_noncohave(void);
void set_mode(void);
void set_input(void);
void set_8_16(int);		/* 0=set to 8 bit, 1=toggle 8/16 */
void set_data_save(void);
void disp_toggle(void);
void normalize_plot(int);	/* 0=clear, 1=15 second, 2=10 min */
void normalize(int);		/* Do the normalization calcs, 256 or 512 */
void do_warm_boot(void);	/* Reset dsp & hardware */
void dofreqplot(int);
void morse(int, int);		/* Morse code routines, (command, ichar) */
void diagnostics(int);
void initialize(void);
void presets(void);
void do_transmit(void);
void do_auto_disp(int k);
void do_gif_file(void);		/* GIF of entire screen */
int  read_config_file(void);
void save_state(void);
void new_band(int);
void rit_toggle(void);
void display_summary(char*, char*);
void print_time(void);
void send_shift_data(unsigned long);  /* Load shift regs and synth from dsp */
void tiny_time(int, int, int);
int  gpixel(int x, int y);
int  get_kbd_int(void);
int  keyident(void);
void get_mem(void);
void quit(void);
void set_agc(int);
void set_cw_speed(void);
void draw_panel(void);
void show_help(void);
void ctrlc_handler(int sig);
void fpe_handler(int sig);
void update_moon(struct tm *ptm, int do_now);
void update_sun(struct tm *ptm, int do_now);
void set_cur_loc(int);
char *left_justify(char *str1, char *str2, int len2, char fill);
void license_info(void);
void license(void);
void warranty(void);
void sun_on(void);
void sun_off(void);
void cold_on(void);
void cold_off(void);

/******************   Global Variables ******************/

char	*big_area_ptr1;
char	*big_area_ptr2;
int		contrast=18;
double		dis_gain=0.0525;
int		bright=64;
double		dis_offset=0.0;
int		power_exp=10;	/* Power of 2 that dsp data is divided by
				   before integration; must be (0, 255).  */
int		scroll_line=0;
/* T/R is controlled either by PTT on rf board or by 'Home'. The following
   three variables allow either but not both at any time.   */
unsigned int	transmit=0;		/* 1=transmit, 0=receive */
unsigned int	ptt=0;			/* PTT: 1=transmit, 0=receive */
unsigned int	home_tr=0;		/* 'Home' key: 1=transmit, 0=receive */
unsigned int	cw_on=0;		/* 0=cw off, 1=cw on */
int	rfgain=0;			/* (0, 6) for 0, -6, ... -36 dB */
int	xmit_power=100;
int	mike_gain=95;
int	afgain=25;
int	cwoffset=600;
int	sidetone_lev=80;
int	fm_squelch=66;
int	noncohave=12;
int	do_window=1;
double  win_adj=0.0;			/* Correctn for Hamming/BH-92 window */
int	fir3[]={0X8, 0X9};		/* Code sent to DSP for two filters */
					/* 0X41 for x64 w/ match600 */
int     do_af_filt=FALSE;		/* 0=no filt, 1=af filt on */
int	af_coef=0;			/* 0=dsp coefs0, 1=coefs1 */
int     do_denoise=FALSE;		/* 0=no denoise, 1=denoise on */
int	denoise_level=50;
int	delay_state=0;			/* 0=No binaural delay */
int	delay_right=9;			/* Power of 2 x 20.8usec on right ch */
int	agc_decay=5;
int	dot_type=2;
int	display_width=FULL_FREQ;
int	bin_low=8;			/* Lowest bin for marker */
int	db_div_set=0;			/* 0=10db/div, 1=5db, 2=2db, 3=1db */
double	kdb_div=3.2;			/* 80 div/256 values */
double	koffset=0.0;			/* Display offset in dB */
int	mode=USB;
int	narrow=1;
int	binaural=0;
int	iq=1;
int	bits_8_16=0;			/* Start 8 bit */
int	input_if_af=0;			/* Start as rf in */
					/* Next var allows for operating with
					   non-standard hardware. Set TRUE only
					   if std hardware available */
int	hardware=TRUE;			/* RF Hardware present */
int	normalize_on=FALSE;		/* 0=off, 1=on, >1 is countdown */
int	auto_disp_on=FALSE;		/* 0=off, 1=on  */
int	rit=0;
int	rit_save=0;
int	rit_on=FALSE;
int	band=1;				/* Band 1 = 144 MHz */
long	tfreqL   =  144200000;
/* The following are personality variables that are set through either the
 * default value, or if present, by the value in the .CFG file.
 * They cannot be set by direct keyboard commands */
struct  palettetype  savpalette;
struct	palettetype  upalette=
			{16,0,16,34,3,25,41,21,29,61,60,52,46,38,54,55,63};
int	startup_detail=1;		/* 0=No detail, 1=Show diagnostics
					   Set =1 in case .cfg missing */
int	waterfall=1;			/* 0=Fancy freq dial, 1=Waterfall */
int	time_type=0;			/* 0=GMT, 1=local */
char	time_str[10]="Z";		/* Change to "PST" etc from .CFG file*/
char	call_str[10]="   ";		/* Enter from .CFG file */
int	quit_save_state=TRUE;		/* TRUE=save state on quit() */
int	beep_on_error=TRUE;		/* When error codes are displayed */
int	error_on_tr=2;			/* Prevent error after T/R change */
struct	transverter_info tvtr[21];

/* Following 3 variables control the synthesizer, dsp 2nd lo and hardware
   setting of the shift registers.  This needs to be collected into
   a structure for neatness, some day!   */
unsigned int	dsp_deltap_lo=0XA6;	/* Phase incr dsp, low 8 bits */
unsigned int	dsp_deltap_hi=0X4A;	/* Phase incr dsp, high 8 bits */
/* shift_reg:  Bit 0=CW on, 1=trans, 2=recv, 3=PA enable, 4=Ant relay transmit,
   High 8 bits come from rfg[i].gainbyte    */
unsigned int	shift_reg=0X04;		/* Low 8 bit hardware control */

int	dsp_status[18];
int	dsp_xmit_status[10];
double	ave_spectral_pwr=0.0;		/* Ave if color agc on */
double	ref_offset=1.0;			/* Correction factor for freq ref */
int	integrate_count=0;		/* Varies from 0 to noncohave/6  */
int	contrast_line_count=0;		/* Display contrast line if >0 */
int	cw_speed=10;			/* Index of speed (0, 25) */
int	display_error_count=0;
int	diag_mode=2;			/* 0=none, 1=text, 2=bargraphs
					   3=Moon/Sun data */
int     do_data_save=0;                 /* by RBR 0=no save to file, 1=save */
unsigned int	com_port;
int	incount;
double	fgain[101];			/* 'Exact' gain value */
int	fgainl[101];			/* Low 8 bits of gain value */
int	fgainh[101];			/*   and high 8 bits */
int	raw_data[515];
double	dnorm[515];
double	normalize_num;
double	pwrdat[515];
double	power10[256];			/* Convert log to power */
FILE*	data_file;
FILE*	config_file;
FILE*   license_file;
FILE*   sun_file; /* RBR */
FILE*   cold_file;
int     sun_han = 0;
int     cold_han = 0;
double  sun_c;
struct  rf_gn  rfg[] = {0.00,  0x88, 100,
			-6.0,  0x48, 94,
			-12.0, 0x28, 88,
			-18.0, 0x18, 82,
			-24.0, 0x14, 76,
			-30.0, 0x12, 70,
			-36.0, 0x11, 64,
			-99.0, 0x00, 00};
double	gain_adj=0.0;			/* dB factor to correct pwr disp */
int     screen_save=TRUE; /* changed by RBR */

/* For moonsun we need the data on the home location and dx.  To allow multiple
 * dx stations there is provision for loc[0] which is always used
 * and up to 9 'dx' locations, one of which could be loc[0] for self echoes.
 * These are normally defined in the .cfg file.
 */
struct	location loc[10];
int	cur_loc=0;
struct	eenv	 cur_env;		/* Cur enviornment:day, nutation, etc*/
int	ut_available=0;			/* 0 if not avail, else 1 */
double	ut=0.0;				/* UT for moonsun calculations */
double	f_eme=144200000.0;		/* Dopplr Fr (audio processing only) */

/* ------------------------------------------------------------------------ */

void main(int argc, char *argv[])
	{
	int grdrv, grmode, colormax, i;
	int cx, new_char, last_char;
	int stop_now;
	int esc_count;
	int y1, y2;
	int check_sum_dsp;		/* Received from dsp */
	int check_sum_pc;		/* Calculated by pc frm data*/ 
	int xmax, ymax;
	char  cc[50], *cptr;
	unsigned int is;
	time_t t0;
	struct tm *tp;
	char *ptc;
	char version[10];

	strcpy(version, "1.60");
	cptr = &cc[0];
	stop_now=FALSE;
	/* Get command line arguments for com port */
	if (argc==1)			/* Should be 2 arguments */
	    {
	    display_summary(argv[0], version);
	    exit(0);
	    }
	sscanf(argv[1], "%d", &com_port);
	if (com_port<1 || com_port>4)
	    {
	    beep(150);
	    printf("COM port must be 1, 2, 3 or 4\n");
	    exit(0);
	    }
	if (com_port==1) com_port=COM1;
	if (com_port==2) com_port=COM2;
	if (com_port==3) com_port=COM3;
	if (com_port==4) com_port=COM4;

	/* NOTE: Read the license before deletion or modification to any
	 * of the following license related code:
	 */
        if (NULL == ( license_file=fopen("GNU_GPL.txt", "r") ))
	    {
	    printf("\nNo GNU_GPL.txt license file was found. This must be in");
	    printf(" the same directory\nas UHFA.EXE file. Exiting program.");
	    exit(1);
	    }
	fclose(license_file);
	license_info();
	for(;;)
	    {
	    if ( kbhit() )
		{		
		cx=getch() & 0X7F;
		if (cx==13)
		    {clrscr(); break;}
		else if (cx=='1')
		    {warranty(); license_info();}
		else if (cx=='2')
		    {license(); license_info();}
		else license_info();
	        }
	    }

	if ((ptc=getenv("TZ")) != 0)
	    {
	    ut_available=1;
	    tzset();
	    printf("\nGMT Conversion from 'SET TZ=%s'", ptc);
	    time(&t0);
	    tp=localtime(&t0);
	    strcpy(cc, asctime(tp));   /* Get copy to rearrange */
	    cc[24]=0;
	    printf("\nComputer time =  %s (local)", cc);
	    tp=gmtime(&t0);
	    strcpy(cc, asctime(tp));
	    cc[24]=0;
	    printf("\n   translates to %s GMT\n\n", cc);
	    }
	else
	    {
	    ut_available=0;
	    printf("\nMoon & Sun data unavailable, needs 'SET TZ=PST8', etc.");
	    time(&t0);
	    tp=localtime(&t0);		/* time_type==0 --> use GMT */
	    strcpy(cc, asctime(tp));   /* Get copy to rearrange */
	    cc[24]=0;
	    printf("\nComputer time=%s (local)", cc);
	    }

	presets();			/* These may be overwritten */
	if (read_config_file() == 0)
	    {
	    printf("\nNo uhfa.cfg present, using default values");
	    }

	get_mem();			/* Memory for spectral disp scroll */
	if (input_if_af==1) tfreqL=0;	/* Prevent inconsistency when in af */

	if(startup_detail)
	   {
	   grdrv=DETECT;	
	   detectgraph(&grdrv, &grmode);

	   printf("Graphics driver= %d, Mode %d\n", grdrv, grmode);
	   if (grdrv != 9)
	      {
	      beep(1000);
	      printf("If driver is not 9 (VGA) & Mode is not 2 (640x480) ");
	      printf("you may encounter\ngraphics driver incompatibilities. ");
	      printf("Hit ENTER to continue\n");
	      }
	   printf("Graphresult= %d,  Hit 'Enter' to proceed", graphresult() );
	   getch();
	   }
	else
	   delay(2000);

	grdrv=9;			/* Only allow VGA 640x480 */
	grmode=2;
	initgraph(&grdrv, &grmode, "");
	strcpy(cptr, getdrivername() );
	xmax=getmaxx() + 1;  ymax = getmaxy() + 1;  /* Global */
	colormax = getmaxcolor();
	is = imagesize(1, 10,  511, 179);
	if (signal(SIGINT, ctrlc_handler)==SIG_ERR)
	    {
	    printf("\nInvalid signal number for Ctrl-C handler\n");
	    exit(0);
	    }
	if (signal(SIGFPE, fpe_handler)==SIG_ERR)
	    {
	    printf("\nInvalid signal number for Floating Point ");
	    printf("error handler\n");
	    exit(0);
	    }
	if(startup_detail)
	    {
	    closegraph();
	    printf("Graphics is %s, Mode %d, %d x %d\n",
		cptr, grmode, xmax, ymax);
	    printf("Bottom image size (170 lines worth) is %u bytes\n",is);
	    printf("Max color is %d\n", colormax);
	    printf("Graphresult= %d\n\n", graphresult() );
	    printf("\nHit 'Enter' (with Scroll Lock off) to proceed.\n");
	    printf("Any other key will give 'key ident' for uhfa_kbd.h.\n");
	    printf("Exit program with Scroll Lock Alt F4. See Readme\n");
	    printf("file for list of keyboard commands.  No mouse is used.\n");
	    for(;;)
		{
		if ( kbhit() )
		    if ( keyident() )  break;
		}
	    initgraph(&grdrv, &grmode, "");
	    }
	getpalette(&savpalette);	/* Palette when we arrived */
	setallpalette(&upalette);
	if ( graphresult()==grError )
		do_error(90, 200);
	setbkcolor(UBLUE);
	draw_grid();
	initialize();
	scroll_line = 0;
        /* sun_file=fopen("sun.txt", "wt");   RBR */
	/* The following is the main control loop for the transceiver */
 pp:	esc_count = 0;
	for(;;)				/* Wait for 1st character */
		{
		if ( ASYNCINSTAT() > 0 )	/* Have ch, see if ESC_CHAR */
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR)  esc_count++;
			 break;			/* Break the for(;;) */
			}
		 while ( kbhit() )		/* Check for commands */
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit) do_transmit();
		    }
		}
	for(;;)			/* Wait here for ESC_CHAR A to come */
		{last_char=new_char;
		 if(ASYNCINSTAT() > 0)
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;	
			 if (esc_count==2)	/* Waiting for ESC_CHAR, 'A' */
				{esc_count=0;	/*    but got 2xESC_CHAR     */
				 goto pp;	/* So try again              */
				}
			 if ((last_char==ESC_CHAR) && (new_char=='A'))
				break;
			}
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit) do_transmit();
		    }
		}
	for(;;)				/* Wait for a buf full */
		{if ( ASYNCINSTAT() > 520 )	/* Almost full */
			 break;
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit) do_transmit();
		    }
		}
	if (--display_error_count == 1)
	    {
	    do_error(9999, 0);		/* Clears any error */
	    display_error_count=0;	/* Don't do it until another error */
	    }
	incount=-1;			/* incount is for diagnostic display */
	i=0;  check_sum_pc=0;  esc_count=0;
	for(;;)	/* Stay here until 512 data points are received */
		{if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 ++incount;
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto skip_char;
				}
			 raw_data[i]=new_char;
			 check_sum_pc += new_char & 0XFF;
			 if (++i > 511)
				{if (esc_count==1) { delay(2); ASYNCIN(); }
				 break;
				}
 skip_char:		}
		}
	for(;;)				/* Get check sum */
		{if ( ASYNCINSTAT() > 0 )
			{check_sum_dsp=ASYNCIN();
/*			 if(check_sum_dsp==ESC_CHAR) {delay(2); ASYNCIN();} */
			 break;
			}
		}
	check_sum_pc &= 0XFF;
	if (error_on_tr > 0)		/* Prevent error after T/R change */
	    error_on_tr--;
	if (check_sum_pc==check_sum_dsp)
	    {
	    update_data();
	    }
	else
	    {
	    if (error_on_tr==0) do_error(1, 300);
	    goto pp;
	    }
	/* If contrast or brightness was recently changed, do color bar: */
	if (--contrast_line_count>0)
		{setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
		 for (i=1; i<16; i++)
		     {setfillstyle(1, i);
		      /* y1=bottom of bar */
		      y1=128-(int)((dis_offset+((double)i)/dis_gain)/2.0);
		      if (y1>127)  y1=127;
		      if (i==1)  y1=127;
		      y2=128-(int)((dis_offset+((double)(i+1))/dis_gain)/2.0);
		      if (y2<1) y2=1;
		      if (i==15)  y2=1;
		      bar(390, y2, 415, y1);
		     }
		 setcolor(UWHITE);
		}
	else
		contrast_line_count=0;

	/* Now get dsp_status bytes */
	i=0; esc_count=0;
	for(;;)
		{if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto dstat1;
				}
			 dsp_status[i] = new_char;
			 if (++i > 17) 
				 break;
 dstat1:			}
		}
	if (check_sum_pc==check_sum_dsp)
	    {
	     if (mode==5) pua43_update(dsp_status[14]);
	     if (mode==4) lhl_update(dsp_status[14]);
	    }
	if(dsp_status[17]==2)		/* Valid status data? */
	    /* See if PTT (not valid w/o hardware) */
	    {if ((dsp_status[16] & 0X10) && hardware)
		{ptt=1;			/* Set to override 'Home' key */
		 t_r_toggle();		/* Go to transmit */
		 do_transmit();		/* And wait here */
		}
	     if (mode!=4) diagnostics(1);	/* Update diagnostics */
	    }
	else
	    do_error(102, 0);	
	print_time();
	goto pp;	
	}				/* End main() */

/*  UPDATE_DATA()  -  Called after 512 data points if check sum is OK.
 *  Converts log to power, adds to power sums and displays on both upper fft
 *  screen as well as lower spectral map.
 */
void update_data(void)
	{
	static time_line_count=0;	/* Decides when to do time print */
	static double auto_disp_base=0;	/* Base value for setting color agc */
	double auto_disp_correct;
	double correction;
 	char cc[50];
	int i;
	int ix;
	int pclr;			/* Color for spectral map, 0 to 7 */
	int biggest;			/* Highest during sweep */
	int biggest_freq;
	int display_data;
	int newy, oldx, oldy;
	double scale;
	double bar_pwr, big_bp;
	time_t tnow;
	struct tm *tmnow;
	double pave, lpave;
	double dbm_level;		/* in i-f, non-FM mode, fft power */
	double fm_log_pwr;		/* Measured log power in FM mode */

	#define BASE_ADJ 126		/* Bottom of spectral display */
	big_bp=0.0;

	/* Convert log raw data to power and add powers for non-coherent
	   integration.  For 8 bit data we use a lookup table, power10[],
	   to do a fast conversion.  Table is too big for 16 bit data.  */
	if (bits_8_16 == 0)		/* If using 8 bit log data */
	    {
	    for (i=0; i<512; i++)
		{
		bar_pwr=power10[ raw_data[i] ];
	 	pwrdat[i] += bar_pwr;	/* Non-coherent integration */
		if (bar_pwr > big_bp)   big_bp=bar_pwr;
		}
	    if (normalize_on > 1) normalize(512);
	    }				/* End 8 bit non-co ave */
	else				/* using 16 bit log data */
	    {
	    for (i=0; i<256; i++)
		{
		/* The following makes the value of 16 bit data the
		    same as that of 8 bit */
		/* Note: LOG_OFFSET_D 12288.0    0.0078125=1/128 */
		bar_pwr=0.0078125*( (double)(raw_data[2*i]+256*raw_data[2*i+1])
		    +LOG_OFFSET_D );
		bar_pwr=pow(10.0, 0.03125*bar_pwr);  /* Same as power10[] */
	 	pwrdat[i] += bar_pwr;	/* Non-coherent integration */
		if (bar_pwr > big_bp)   big_bp=bar_pwr;
		}
	    for (i=256; i<512; i++)	/* Dummy out unused spectrum */
		pwrdat[i] = 0.00001;
	    if (normalize_on > 1) normalize(256);
	    }				/* End 16 bit non-co ave */

	/* After noncohave/6 times we are ready to update all spectral
	   displays:						*/
	if ( ++integrate_count>=(noncohave/6) )	/* If time for display */
	    {
	    auto_disp_correct=0.0;
	    if (auto_disp_on==TRUE && (auto_disp_correct<-10.0 |
			auto_disp_correct>10.0) )
		do_error(70, 10);
	    if (auto_disp_on==-1)	/* Establish base average level */
		{
		auto_disp_on=TRUE;
		auto_disp_base=0.0;
		for (i=0; i<(512-display_width*256); i++)
		    auto_disp_base += pwrdat[i];
		}
	    if ( (auto_disp_on==TRUE) && (auto_disp_base>1.0e-100) )
		{
		for (i=FLOW; i<(512-display_width*256); i++) /* Cur ave lev */
		    auto_disp_correct += pwrdat[i];
		/* For display, remove dc term: */
		ave_spectral_pwr=auto_disp_correct - pwrdat[0] - pwrdat[1];
		auto_disp_correct /= auto_disp_base;  /* Make rel to base */
		if (auto_disp_correct>1.0e-100)
		    auto_disp_correct = 10.0*log10(auto_disp_correct);
		else
		    auto_disp_correct=0.0;
		}
	    integrate_count=0;
	    biggest=0; biggest_freq=0;
	    prep_display();
	    /* scale accounts for non coherent averaging */
	    scale=10.0*log10( (double)noncohave );
	    oldx=0;
	    oldy=128;
	    /* -3*power_exp+30 is a correction for bit data scaling in DSP */
	    correction = -6.0-koffset+scale+auto_disp_correct
		-3.0*(double)power_exp+30.0;
	    if(display_width==FULL_FREQ)
		{
		for (i=0; i<512; i++)
		    {
		    display_data = (int)
			( kdb_div*(10.0*log10(pwrdat[i]) - dnorm[i]
			     - correction) );
		    setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
		    if (dot_type==0)
			putpixel(i, BASE_ADJ-(display_data >> 1), UWHITE);
		    if (dot_type==1)
			point1(i, BASE_ADJ-(display_data >> 1) );
		    if (dot_type==2)
			{
			newy = BASE_ADJ-(display_data >> 1);
			line(oldx, oldy, i, newy );
			oldx=i;
			oldy=newy;
			}
		    if (waterfall)
			{
			setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
			pclr=(int)
			    (dis_gain*((double)display_data - dis_offset));
			if (pclr>15) pclr=15;
			if (pclr<1) pclr=1;
			putpixel(i, scroll_line, pclr);
			}
		    /* bin_low*9.375 is lowest freq considered for marker*/
		    if (display_data > biggest  &&  i>bin_low )
			{
			biggest=display_data;
			biggest_freq=i;
			}
		    }		/* End 512 times */
		}		/* End FULL_FREQ display */
	    else 		/* if (display_width==HALF_FREQ) */
		{
		for (i=0; i<256; i++)
		    {
		    display_data = (int)
			( kdb_div*(10.0*log10(pwrdat[i]) - dnorm[i]
			     - correction) );
		    setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
		    if (dot_type==0)
			   putpixel(2*i,BASE_ADJ-(display_data >> 1), UWHITE);
			   putpixel(2*i+1,BASE_ADJ-(display_data >> 1),UWHITE);
		    if (dot_type==1)
			point1(2*i, BASE_ADJ-(display_data >> 1) );
		    if (dot_type==2)
			    {
			    newy = BASE_ADJ-(display_data >> 1);
			    line(oldx, oldy, 2*i, newy );
			    oldx=2*i;
			    oldy=newy;
			    }
		    if (waterfall)
			{
			setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
			pclr=(int)
			    (dis_gain*((double)display_data - dis_offset));
			if (pclr>15) pclr=15;
			if (pclr<1) pclr=1;
			putpixel(2*i,   scroll_line, pclr);
			putpixel(2*i+1, scroll_line, pclr);
			}
			/* bin_low*9.375 is lowest freq for marker*/
			if (display_data>biggest  &&  i>bin_low )
			    {
			    biggest=display_data;
			    biggest_freq=i;
			    }
		    }			/* End of 256 times */
	        }
	    /* Add marker at highest point in upper spectral display */
	    setcolor(ULIGHTRED);
	    setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
	    if (biggest > 250) biggest=250;
	    point5((1+display_width)*biggest_freq, BASE_ADJ-(biggest >> 1) );

	    /* Add red line at cw offset */
	    i=(1+display_width)*(int)((double)cwoffset/9.375);
	    if(mode==CW) line(i, 120, i, 128);
	    setcolor(UWHITE);
	    time(&tnow);
	    if (time_type)
		tmnow = localtime(&tnow);	/* time_type==1 for local */
	    else
		tmnow = gmtime(&tnow);		/*          ==0 for gmt */
	    if (waterfall && time_line_count++ >= 20)
		{
		time_line_count = 0;
		if (scroll_line > 9)	/* Not above box 2 */
		    {
		    setviewport(X2L+1, Y2T+scroll_line-7,
			X2L+33, Y2T+scroll_line+1, DO_CLIP);
		    clearviewport();
		    setcolor(UWHITE);
		    tiny_time( (100*tmnow->tm_hour+tmnow->tm_min), 1, 8);
		    }
		}

	    if (do_data_save)
	        {
	        if (NULL == ( data_file=fopen("C:\UHFA_1.DAT", "a+t") ))
		  {
		  printf("Unable to open output file\n");
		  sound(500);
		  delay(200);
		  nosound();
		  }
		else
		  {
		  setviewport(0,80,101,90,DO_CLIP);
		  clearviewport();
		  setcolor(ULIGHTRED);
		  settextstyle(0, 0, 1);
		  settextjustify(1,2);
		  moveto(50,1);  outtext("RECORDING");
	          fprintf(data_file,
		    "HA %d %d %d %02d%02d %d\n",
		    tmnow->tm_year, tmnow->tm_mon + 1, tmnow->tm_mday,
		    tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec);
	          fprintf(data_file,
		    "HB %d %d %d %d %d %d %d %d %d %d\n",
		    0, rfgain, noncohave, do_window,
		    256*dsp_deltap_hi+dsp_deltap_lo,
		    4*mode+16*bits_8_16+64*input_if_af, 
		    0, 0, 0, 16);		/* Spares */
	    	  for (i=0; i<32; i++)
		    fprintf(data_file,
		      "%E %E %E %E %E %E %E %E\n",
		      pwrdat[8*i],pwrdat[8*i+1],pwrdat[8*i+2],pwrdat[8*i+3],
		      pwrdat[8*i+4],pwrdat[8*i+5],pwrdat[8*i+6],pwrdat[8*i+7]);
		  fclose(data_file);
		  clearviewport();
		  setcolor(UBEIGE);
		  moveto(50,1);  outtext("RECORDING");
	          }
	        }			/* End data save to disk file */

	    /* S-Meter bar graph: */
	    setviewport(0, S_METER_POS-20, 127, S_METER_POS-10, DO_CLIP);
	    if (mode==FM)
		{
		/* In following, the value in status[5] and [6] is
		  1000*log10(pwr) and the bargraph routine wants
		  abs(max volts).  Variable gain_adj comes
		  from uhfa.cfg file and allows compensation for gain variation
		  between units. It is in dB and can be + or -.
		  The 4.0* factor moves the display over 2 bars so that a max
		  signal of about -54 dBm (rf gain full up) lights all s-meter
		  bars  */
 		ix = 256*dsp_status[6]+dsp_status[5]+(int)(100.0*gain_adj)
		    -(int)(100.0*(rfg[rfgain].dbgain+tvtr[band].gain_db));
				  /* This last was i-f pwr in db*100 */
		if (dsp_status[17]==2)		/* If status data good */
		    bar_graph( 20, 0,
		      (unsigned int)(4.0*sqrt(
		      pow(10.0,((double)(3703+ix))/1000.0))));
		/* 93.00 relates to the gain of the radio */
		fm_log_pwr = -93.00+((double)ix)/100.0;
		}
	    else if (input_if_af==0 && mode!=FM)
		{
		dbm_level =  -161.8-rfg[rfgain].dbgain
		    +10.0*log10(pwrdat[biggest_freq])+gain_adj+win_adj
		    +3.0*(double)(power_exp-10)
		    -tvtr[band].gain_db;
		bar_graph( 20, 0,
		    (unsigned int)pow(10.0,(dbm_level+156.0)/20.0) );
		}
	    else				/* If af input, not FM */
		bar_graph(20, 0, (unsigned int)(3.0*sqrt(big_bp)) );

	    /* Display power below bar graph: */
	    setviewport(1,S_METER_POS+7, 127, S_METER_POS+29, DO_CLIP);
	    clearviewport();
	    settextjustify(1,2);
	    settextstyle(0, 0, 1);
	    setcolor(UWHITE);
	    if (input_if_af==0)
		sprintf(cc, "%.1lf dBm", dbm_level);	/* RF mode */
	    else
		sprintf(cc, "%.1lf dB", ((double)biggest)/kdb_div);
	    if (mode==FM)		/* If FM, overwrite pwr, rf or af */
		sprintf(cc, "%.2lf dBm", fm_log_pwr);
            moveto(64, 0);
	    outtext(cc);

	    sprintf(cc, "%d Hz", (int)(9.375*(double)biggest_freq));
            moveto(64, 10);
	    outtext(cc);

	    /* Unless diagnostics is blank, print the relative average power
	       over freq range (187, 2400 Hz).			*/
	    setviewport(580, 470, 639, 479, DO_CLIP);
	    clearviewport();
	    if (diag_mode>0)
		{
		pave = 0.0;
		for  (i=20; i<255; i++)
		    pave += pwrdat[i];
		/* 3*power_exp+30 corrects for scaling in dsp for 8 bit data */
		if (bits_8_16==0)		/* If using 8 bit log data */
		    lpave = 10.0*log10(pave)-23.692+3.0*(double)power_exp-30.0;
		else
		    lpave = 10.0*log10(pave)-23.692;
		settextjustify(RIGHT_TEXT, TOP_TEXT);
		settextstyle(0, 0, 1);
		setcolor(UWHITE);
		sprintf(cc, "%.2lfdB", lpave);
		moveto(58, 0);
		outtext(cc);

                /*  RBR this is where you output the SUN data */
                if (sun_han ==1)
                     {
                  sun_c=atof(cc);
                  fprintf(sun_file,"%4.2f",sun_c);
                  fputs("\n",sun_file);
                  setviewport(0,80,105,90,DO_CLIP);
		  clearviewport();
		  setcolor(ULIGHTRED);
		  settextstyle(0, 0, 1);
		  settextjustify(1,2);
                  moveto(50,1);  outtext("RECORD SUN");
                  clearviewport();
		  setcolor(UBEIGE);
                  moveto(50,1);  outtext("RECORD SUN"); 
                     }
                if (cold_han == 1)
                     {
                     sun_c=atof(cc);
                  fprintf(cold_file,"%4.2f", sun_c,"  ");
                  fputs("\n",cold_file);
                  setviewport(0,80,105,90,DO_CLIP);
		  clearviewport();
		  setcolor(ULIGHTRED);
		  settextstyle(0, 0, 1);
		  settextjustify(1,2);
                  moveto(50,1);  outtext("RECORD COLD");
		  clearviewport();
		  setcolor(UBEIGE);
                  moveto(50,1);  outtext("RECORD COLD"); 
                      }


                }
                 
	    /* Dump non-coherent integrator to start a new line */
	    for  (i=0; i<512; i++)
		pwrdat[i]=0;		/* Dump integrator */
                }                       /* End display update */
                }



/* print_big(ch, pos, num) prints a double size character in the top
 * box.  The position can be specified as 0 to 27, or by setting pos=255
 * it will be printed in the next available position.  Scrolls to the left
 * when full, Times Sq style (if pos=255.)  If pos=-1 then initialize.
 * Also prints a 2 digit number below the character (future feature.)
 */

void print_big(char ch, int pos, int num)
	{static int print_pos;
	 static char c[28];
	 int i;

	 if (pos==-1)
	    {for(i=0; i<28; i++)  c[i]=32;
	     c[27]=0;
	     print_pos=0;
	    }
	 else if(pos==255)		/* In scroll mode */
	    {if (print_pos==27)		/* In scroll position */
		{for (i=0; i<26; i++)  c[i]=c[i+1];
		 c[26]=ch;
		}
	     else
		{c[print_pos++]=ch;
		}
	    }
	 else
	    c[pos]=ch;
	 setviewport(129, 12, 638, 49, DO_CLIP);
	 clearviewport();
	 settextstyle(0, 0, 2);
	 settextjustify(0,2);
	 moveto(3,10);
	 outtext(c);
	 i=num;  /* Temporary to stop warning from compiler */
/*	 if (num>=0)
	    {tiny_char(num/10, 15*pos, 35);
	     tiny_char(num-(num/10), 6+15*pos, 35);
	    }*/
	}


/* print_medium(ch, pos) prints a standard size character in the top
 * box.  The position can be specified as 0 to 100, or by setting pos=255
 * it will be printed in the next available position.  Scrolls to the left
 * when full, Times Sq style (if pos=255.)  If pos=-1 then initialize.
 */
void print_medium(char ch, int pos)
	{
	 #define LINE_LEN 60
	 static int print_pos;
	 static char c[3*LINE_LEN];
	 char save_ch;
	 int i;

	 if (pos==-1)			/* Code for initialize buffer */
	    {for(i=0; i<3*LINE_LEN; i++)  c[i]=32;  /* Blank */
	     c[3*LINE_LEN-1]=0;
	     print_pos=0;		/* Start at first position */
	    }
	 else if(pos==254)		/* Code for backspace */
	    {
	    print_pos--;
	    c[print_pos] = 32;		/* Erase last char */
	    }
	 else if(pos==255)		/* In scroll mode */
	    {if (print_pos==3*LINE_LEN-1)	/* In scroll position */
		{for (i=0; i<3*LINE_LEN-2; i++)  c[i]=c[i+1];
		 c[3*LINE_LEN-2]=ch;
		}
	     else
		{c[print_pos++]=ch;
		}
	    }
	 else
	    {
	    c[pos]=ch;
	    }
	 if (waterfall)
	     setviewport(129, 10, 638, 49, DO_CLIP);
	 else
	     setviewport(X2L+1, Y2T+59, X2R-1, Y2T+98, DO_CLIP);
	 clearviewport();
	 settextstyle(0, 0, 1);
	 settextjustify(LEFT_TEXT, TOP_TEXT);
	 moveto(3, 8);
	 save_ch = c[LINE_LEN];  c[LINE_LEN] = 0;
	 outtext(c);
	 c[LINE_LEN] = save_ch;
	 save_ch = c[2*LINE_LEN];  c[2*LINE_LEN] = 0;
	 moveto(3, 18);
	 outtext(&c[LINE_LEN]);
	 c[2*LINE_LEN] = save_ch;
	 moveto(3, 28);
	 outtext(&c[2*LINE_LEN]);
	}

void prep_display(void)
	{int i, x, y;
	/* If not waterfall, put data below frequency dial */
	if ((mode<4 && mode!=CW) || !waterfall)
	    {
	    setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
	    clearviewport();
	    setcolor(UBEIGE);
	    for(i=1; i<=8; i++)
		{
		y = i*16;
		line(0, y, 639, y);	/* Horizontal grid lines */
		}
	    for(i=1; i<=9; i++)
		{
		x = (int)(53.229 * (double)i);
		line(x, 1, x, 128);		/* Vertical lines */
		}
	    }
	else		/* For CW & data modes, top of spectral disp */
	    {
	    setviewport(X1L+1, Y1T+49, X1R-1, Y1B-1, DO_CLIP);
	    clearviewport();
	    setcolor(UBEIGE);
	    for(i=0; i<=5; i++)
		{
		y = i*16;
		line(1, y, 639, y);	/* Horizontal grid lines */
		}
	    for(i=1; i<=9; i++)
		{
		x = (int)(53.229 * (double)i);
		line(x, 1, x, 80);		/* Vertical lines */
		}
	    setviewport(X1L+1,2,X1R-1,9,DO_CLIP); /* Tidy up top stripe */
	    clearviewport();
	    }

	set_disp_offset(0);		/* Update display, no dsp command */
	set_db_div(0);			/* Update display only */

	setcolor(UWHITE);
	setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
	if(waterfall && ++scroll_line > 321)
		{
		scroll_line=312;
		getimage(0,  10, 510, 170, big_area_ptr1);
		getimage(0, 171, 510, 321, big_area_ptr2);
		setviewport(X2L+1, Y2B-30, X2R-1, Y2B-10, DO_CLIP);
		clearviewport();
		setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
		putimage(0,   0, big_area_ptr1, COPY_PUT);
		putimage(0, 161, big_area_ptr2, COPY_PUT);
		}
	}

void normalize(int npoints)
	{
	int i;
	double dnorm_big;

	/* Accumulate noise power to find frequency response */
	if (npoints==512)
	    /* For 512 points of 8 bit data from dsp: */
	    for (i=0; i<npoints; i++)
	    	dnorm[i] += (double)raw_data[i];
	else
	    /* For 256 points of 16 bit data from dsp: */
	    for (i=0; i<npoints; i++)
		dnorm[i] += (double)
		    ( (raw_data[2*i]+(raw_data[2*i+1]<<8)) >>7);

	if (--normalize_on == 1)	/* Used as flag and as a counter */
	    {
	    setviewport(0,80,101,90,DO_CLIP);
	    clearviewport();
	    settextstyle(0, 0, 1);
	    settextjustify(1,2);
	    moveto(50,1);
	    if (normalize_num<100.0) outtext("Fast Norm");
	    else outtext("Prec Norm");

	    dnorm_big=-32768.0;
	    for (i=0; i<npoints; i++)
		{
		dnorm[i]=dnorm[i]/(3.2*normalize_num);
		if (dnorm[i]>dnorm_big)  dnorm_big=dnorm[i];
		}
	    for (i=0; i<npoints; i++)
		{
		dnorm[i] -= dnorm_big;
		}
	    }
	}

/* do_transmit()  -  Here because the transmit/receive toggle was put into
 * the transmit position.  We will stay here until receive is again commanded.
 */
void do_transmit(void)
	{
	int i, alternate;
	int stop_now;
	int new_char, last_char;
	int esc_count;
	int check_sum_dsp;		/* Received from dsp */
	int check_sum_pc;		/* Calculated by pc frm data*/ 

	stop_now=FALSE;
	alternate=0;

 ppt:	/* First clear excess status input */
	while (ASYNCINSTAT() >60 )
		ASYNCIN();
	for(;;)				/* Wait for ESC_CHAR */
		{if ( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR)  esc_count++;
			 break;
			}
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit==0)  return;	/* Back to receive */
		    }
	    	 if (mode==CW)  morse(SEND_CW, 0); /* Send ch, if available */
		}
	if(alternate++ <= 5)   goto ppt;
	alternate=0;
	for(;;)				/* Wait for 'T' */
		{last_char =new_char;
		 if(ASYNCINSTAT() > 0)
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;	
			 if (esc_count==2)	/* Waiting for ESC_CHAR, 'T' */
				{esc_count=0;	/*    but got 2xESC_CHAR     */
				 goto ppt;	/* So try again              */
				}
			 if ((last_char==ESC_CHAR) && (new_char=='T'))
				break;
			}
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit==0)  return;	/* Back to receive */
		    }
	    	 if (mode==CW)  morse(SEND_CW, 0); /* Send ch, if available */
		}
	for(;;)				/* Wait for a buf full */
		{if ( ASYNCINSTAT() > 29 )	/* Almost full */
			 break;
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit==0)  return;	/* Back to receive */
		    }
	    	 if (mode==CW)  morse(SEND_CW, 0); /* Send ch, if available */
		}
	incount=-1;
	i=0;  check_sum_pc=0;  esc_count=0;
	for(;;)				/* Get xmit_status */
		{if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 ++incount;
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto t_skip_char;
				}
			 dsp_xmit_status[i]=new_char;
			 check_sum_pc += new_char & 0XFF;
			 if (++i > 8)
				{if (esc_count==1) ASYNCIN();
				 break;
				}
 t_skip_char:		}
		}
	for(;;)				/* Get check sum */
		{if ( ASYNCINSTAT() > 0 )
			{check_sum_dsp=ASYNCIN();
			 if (check_sum_dsp==ESC_CHAR) ASYNCIN();
			 break;
			}
		}
	check_sum_pc &= 0XFF;
	if (check_sum_pc==check_sum_dsp)  
		{
		}
	else /*   do_error(111, 30)  */ ;
	
	i=0; esc_count=0;
	for(;;)				/* Wait for dsp_status[] */
		{if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto t_dstat1;
				}
			 dsp_status[i] = new_char;
			 if (++i > 17) 
				{if (esc_count==1) ASYNCIN();
				 break;
				}
 t_dstat1:		}
		}
	if(dsp_status[17]==2)		/* Valid status data? */
	    {if  ( ((dsp_status[16] & 0X10)==0) && ptt)  /* PTT to receive */
		{ptt=0;
		 t_r_toggle();		/* Back to receive */
		 return;
		}
	     diagnostics(1);		/* Update diagnostics */
	    }
	else
	    do_error(103, 0);
	print_time();
	goto ppt;	
	}
	
void print_stars(void)
	{
	 unsigned int cnt;
	 for(cnt = 0; cnt < 79; cnt ++)
		  printf("*");
	 }

int do_cmd(void)
	{
	char cc[15];
	unsigned int ichr, stat;
	long lkey;
	int irtn;

	stat = _bios_keybrd(_KEYBRD_SHIFTSTATUS);
	/* Make left and right shift keys equals (bits 0 and 1) and
	 * mask out Caps Lock, Num Lock and Insert (bits 5, 6, and 7).     */
	if (stat&2)   stat |= 1;
	stat &= 0X1D;
	ichr = _bios_keybrd(_KEYBRD_READ);
	lkey = (long)ichr + (((long)stat)<<16);
	irtn=0;

	switch (lkey) {
		case CW_BUF_CLR: {
			morse(INIT_CW, 0);
			print_medium(0, -1);
			break; }
		case MIKE_GAIN_M: {
			if (--mike_gain<0) mike_gain++;
			set_mike_gain();
                sun_han=0;
                fclose(sun_file);
                cold_han=1;
                fclose(cold_file);
                       {
                       cold_file=fopen("cold.dat", "wt");  /* RBR */
                       }
			break; }
		case MIKE_GAIN_P: {
			if (++mike_gain>100) mike_gain--;
			set_mike_gain();
                cold_han=0;
                fclose(cold_file);
                        setviewport(0,80,105,90,DO_CLIP);
                        clearviewport();
                	break; }
		case XMT_POWER_M: {
			if (--xmit_power<0) xmit_power++;
			set_xmit_power();
			break; }
		case XMT_POWER_P: {
			if (++xmit_power>100) xmit_power--;
			set_xmit_power();
			break; }
		case CW_SPEED_M: {
			if (--cw_speed<0) cw_speed++;
			set_cw_speed();
			break; }
		case CW_SPEED_P: {
			if (++cw_speed>25) cw_speed--;
			set_cw_speed();
			break; }
		case T_R_TOG: {
			if (ptt) break;	/* No response if ptt in transmit */
			t_r_toggle();
			break; }
		case FREQ_M1: {
			if (mode != FM) tfreqL -= 1;
			new_freq();
			break; }
		case FREQ_P1: {
			if (mode != FM) tfreqL += 1;
			new_freq();
			break; }
		case FREQ_M10: {
			if (mode == FM) tfreqL -= 5000;
			else tfreqL -= 10;
			new_freq();
			break; }
		case FREQ_P10: {
			if (mode == FM) tfreqL += 5000;
			else tfreqL += 10;
			new_freq();
			break; }
		case FREQ_M100: {
			if (mode == FM) tfreqL -= 5000;
			else tfreqL -= 100;
			new_freq();
			break; }
		case FREQ_P100: {
			if (mode == FM) tfreqL += 5000;
			else tfreqL += 100;
			new_freq();
			break; }
		case FREQ_M1K: {
			if (mode == FM) tfreqL -= 5000;
			else tfreqL -= 1000;
			new_freq();
			break; }
		case FREQ_P1K: {
			if (mode == FM) tfreqL += 5000;
			else tfreqL += 1000;
			new_freq();
			break; }
		case FREQ_M10K: {
			tfreqL -= 10000;
			new_freq();
			break; }
		case FREQ_P10K: {
			tfreqL += 10000;
			new_freq();
			break; }
		case FREQ_M100K: {
			tfreqL -= 100000;
			new_freq();
			break; }
		case FREQ_P100K: {
			tfreqL += 100000;
			new_freq();
			break; }
		case FREQ_M1MHZ: {
			tfreqL -= 1000000;
			new_freq();
			break; }
		case FREQ_P1MHZ: {
			tfreqL += 1000000;
			new_freq();
			break; }
		case RIT_M_C:
		case RIT_M_L: {
			if(rit_on) 
			    {
			    rit -= 10;
			    new_freq();
			    }
			break; }
		case RIT_P_C:
		case RIT_P_L: {
			if(rit_on) 
			    {
			    rit += 10;
			    new_freq();
			    }
			break; }
		case RIT_Z_C:
		case RIT_Z_L: {
			rit = 0;
			new_freq();
			break; }
		case RIT_O_C:
		case RIT_O_L: {
			rit_toggle();
			new_freq();
			break; }
		case AF_GAIN_M: {
			--afgain;
			if (--afgain<0) afgain=0;
			set_afgain();
			break; }
		case AF_GAIN_P: {
			++afgain;
			if (++afgain>100) afgain=100;
			set_afgain();
			break; }
		case AF_GAIN_M1: {
			if (--afgain<0) afgain=0;
			set_afgain();
			break; }
		case AF_GAIN_P1: {
			if (++afgain>100) afgain=100;
			set_afgain();
			break; }
		case RF_GAIN_M: {
			if (++rfgain>6) rfgain--;
			set_rfgain();
			break; }
		case RF_GAIN_P: {
			if (--rfgain<0) rfgain++;
			set_rfgain();
			break; }
		case FM_SQUELCH_M_C:
		case FM_SQUELCH_M_L: {
			if (--fm_squelch<0) fm_squelch++;
			set_fm_squelch();
                cold_han=0;
                fclose(cold_file);
                sun_han=1;
                fclose(sun_file);
                        {
                        sun_file=fopen("sun.dat", "wt");  /* RBR */
                        }
        		break; }
		case FM_SQUELCH_P_C:
		case FM_SQUELCH_P_L: {
			if (++fm_squelch>99) fm_squelch--;
			set_fm_squelch();
                sun_han=0;
                fclose(sun_file);  /* RBR */
                        setviewport(0,80,105,90,DO_CLIP);
                        clearviewport();
                        break; }
		case WINDOW_T_C:
		case WINDOW_T_L:  {
			if (++do_window>2)   do_window=1;
			set_window();
			break; }
		case AF_FILTER_T:  {
			do_af_filt ^= 1;
			set_affilt_sw();
			break; }
		case AF_FILTER_S:  {
			af_coef ^= 1;
			set_affilt_select();
			break; }
		case DENOISE_T:  {
			select_denoise();
			break; }
		case DELAY_T_C:
		case DELAY_T_L:  {
			set_delay_right(0);
			break; }
		case DELAY_M: {
			if (--delay_right<1) delay_right=1;
			set_delay_right(delay_right);
			break; }
		case DELAY_P: {
			if (++delay_right>10) delay_right=10;
			set_delay_right(delay_right);
			break; }
		case NC_AVE_M:  {
			noncohave-=6;
			if (noncohave<6) noncohave=6;
			set_noncohave();
			break; }
		case NC_AVE_P:  {
			noncohave+=6;
			if (noncohave>10002) noncohave=10002;
			set_noncohave();
			break; }
		case RECORD_C:
		case RECORD_L: {
			set_data_save();
			break; }
		case GIF_C:
		case GIF_L: {
			do_gif_file();
			break; }
		case DOT_TYPE:  {	/* Display dot */
			set_dot();
			break; }
		case CWOFFSET_M: {
			if (cwoffset<100 || mode!=CW) break;
			cwoffset=cwoffset-10;
			set_cwoffset();
			break; }
		case CWOFFSET_P: {
			if (cwoffset>990 || mode!=CW) break;
			cwoffset=cwoffset+10;
			set_cwoffset();
			break; }
		case SIDETONE_LEV_M: {
			if (--sidetone_lev<0) sidetone_lev++;
			set_cwoffset();
			break; }
		case SIDETONE_LEV_P: {
			if (++sidetone_lev>100) sidetone_lev--;
			set_cwoffset();
			break; }
		case DENOISE_M: {
			denoise_level--;
			if (denoise_level<0)
				denoise_level=0;
			set_denoise();
			break; }
		case DENOISE_P: {
			denoise_level++;
			if (denoise_level>100)
				denoise_level=100;
			set_denoise();
			break; }
		case AGC_DECAY_M: {
			if (mode==FM)  break;
			agc_decay--;
			if (agc_decay<0)
				agc_decay=0;
			set_agc(agc_decay);
			break; }
		case AGC_DECAY_P: {
			if (mode==FM)  break;
			agc_decay++;
			if (agc_decay>10)
				agc_decay=10;
			set_agc(agc_decay);
			break; }
		case BITS_8_16: {	/* Toggle 8/16 bit mode from dsp */
			set_8_16(1);
			break; }
		case 0X1333C:	{	/* Previous location for EME */
			set_cur_loc(0);
			break; }
		case 0X1343E:	{	/* Next location for EME */
			set_cur_loc(1);
			break; }
		}
	/* Be careful that any key in the following must NOT also be above */
	if (transmit==0)		/* Keys available only in receive */
	    {switch(lkey){
		case HELP: {
			show_help();
			break;  }
/*		case AF_FILTER_DESIGN:  {
			filter_design();
			break;  } 		*/
		case MODE_S_C:
		case MODE_S_L:  {
			if (++mode>3)   mode=0;  /* 5 to include data modes */
			/* In audio, USB only to avoid inversion of tuning: */
			if (input_if_af && mode==LSB) ++mode;
			set_mode();
			break; }
		case NARROW_T:  {
			if(input_if_af) break;
			narrow ^= 1;
			set_mode();
			break; }
		case BINAURAL_T:  {
			if(input_if_af) break;
			binaural ^= 1;
			set_mode();
			break; }
		case IQ_T:  {
			if(input_if_af) break;
			iq ^= 1;
			set_mode();
			break; }
		case CONTRAST_M: {
			if (--contrast<0) contrast++;
			set_contrast();
			break; }
		case CONTRAST_P: {
			if (++contrast>100) contrast--;
			set_contrast();
			break; }
		case BRITE_M: {
			if (--bright<0) bright++;
			set_brightness();
			break; }
		case BRITE_P: {
			if (++bright>100) bright--;
			set_brightness();
			break; }
		case DIAG_S:  {		/* Toggle Diagnostics */
			diagnostics(0);
			break; }
		case NORM_QCK_C:
		case NORM_QCK_L:  {	/* Toggle Fast normalize */
			if (normalize_on)  normalize_plot(0);
			else               normalize_plot(1);
			break; }
		case NORM_PRC_C:
		case NORM_PRC_L:  {	/* Toggle Precision normalize */
			if (normalize_on)  normalize_plot(0);
			else               normalize_plot(2);
			break; }
		case COLOR_AGC_C:
		case COLOR_AGC_L:  {	/* Toggle Color AGC */
			if (auto_disp_on)  do_auto_disp(0);
			else               do_auto_disp(1);
			break; }
		case DISP_WIDTH:  {	/* Display width */
			display_width ^= 1;
			set_display_width();
			break; }
		case DISP_T_C:
		case DISP_T_L:	{	/* Toggle waterfall display */
			waterfall ^= 1;
			disp_toggle();
			break; }
		case OFFSET_M: {
			koffset -= 1.0;
			if (koffset<-50.0) koffset=-50.0;
			set_disp_offset(1);
			break; }
		case OFFSET_P: {
			koffset += 1.0;
			if (koffset>50.0) koffset=50.0;
			set_disp_offset(1);
			break; }
		case DB_DIV: {
			db_div_set += 1;
			if (db_div_set>3)  db_div_set=0;
			set_db_div(1);
			break; }
		case DO_QUIT: {		/* Exit to DOS */
			irtn=1;
			break; }
		case WARM_BOOT: {	/* Just in case...  */
			do_warm_boot();
			break; }
		case INPUT_S_C:
		case INPUT_S_L: {
			input_if_af ^= 1;	/* Toggle input */
			set_input();
			break; }
		case BAND_50_A:   { new_band(0);  break; }
		case BAND_144_B:  { new_band(1);  break; }
		case BAND_220_C:  { new_band(2);  break; }
		case BAND_432_D:  { new_band(3);  break; }
		case BAND_902_9:  { new_band(4);  break; }
		case BAND_1296_E: { new_band(5);  break; }
		case BAND_2304_F: { new_band(6);  break; }
		case BAND_3456_G: { new_band(7);  break; }
		case BAND_5760_H: { new_band(8);  break; }
		case BAND_10K_I:  { new_band(9);  break; }
		case BAND_24K_J:  { new_band(10);  break; }
		case BAND_47K_K:  { new_band(11);  break; }
		case BAND_75K_L:  { new_band(12);  break; }
		case BAND_119K_M: { new_band(13);  break; }
		case BAND_145K_N: { new_band(14);  break; }
		case BAND_241K_O: { new_band(15);  break; }
		case BAND_MEMO_1: { new_band(16);  break; }
		case BAND_MEMO_2: { new_band(17);  break; }
		case BAND_MEMO_3: { new_band(18);  break; }
		case BAND_MEMO_4: { new_band(19);  break; }
		case BAND_MEMO_5: { new_band(20);  break; }
		default: break;
		}
	    }
	if (mode==CW)
	   {
	   ichr=ichr & 0XFF;
	   /* Not scrl, ctrl, alt, but either alphnumeric or Back Space */
	   if ((!(lkey & 0X1C0000)) && (((ichr>=32) && (ichr<127)) || ichr==8))
	       morse(ADD_CH, ichr);
	   }
	return(irtn);
	}

/* t_r_toggle()   - Responds to hitting t/r toggling key.  Adds sequencing
 * for external antenna relays and power amps. Bits are controlled by
 * uint shift_reg with  0=CW on, 1=trans, 2=recv, 3=PA enable,
 * 4=Ant relay transmit.
 */
void t_r_toggle(void)
	{char cc[10];
	static unsigned int save_rfgain;
	
/*	if (input_if_af==1)  return; */	/* Non-operative in audio mode */

	cw_on = 0;			/* Start with cw key up */
	transmit ^= 1U;			/* Toggle transmit/receive */

	if (transmit)
		{
		save_rfgain=rfg[rfgain].gainbyte;
	 	rfg[rfgain].gainbyte=0;	/* Disconnect receiver from antenna */
		setcolor(ULIGHTRED);
		sprintf(cc, "%s", "TRANSMIT");
/* Bob-There aer 3 new_freq()'s - Is one at beginning adequate?? */
		new_freq();		/* T & R freq may be different */
		if (hardware)
		    {  
		    shift_reg = 0X0010;    /*(shift_reg & 0X00E0) | 0X0010;*/
		    send_shift_data(R10MHZ); /* Ant ry to transmit, recv off */
		    beep(50);
		    shift_reg=0X0018;    /* |= 0X0008;*/
		    send_shift_data(R10MHZ);	/* Enable power amp */
		    delay(20);
		    shift_reg=0X001A;    /*  |= 0X0002; */
		    send_shift_data(R10MHZ);	/* Hardware into transmit */
		    delay(25);
		    }
		ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		ASYNCOUT(1); ASYNCOUT(2);     /* 1 = transmit | (cw_on<<4)  */ 
		ASYNCOUT( fgainl[xmit_power] ); ASYNCOUT( fgainh[xmit_power] );
		}
	else
		{
		/* Get back out of transmit to allow receiving command */
		ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		ASYNCOUT(0); ASYNCOUT(2);     /* 0 = transmit | (cw_on<<4)  */ 
		ASYNCOUT( fgainl[xmit_power] ); ASYNCOUT( fgainh[xmit_power] );
		setcolor(UWHITE);
		if(input_if_af)		/* If audio mode */
		    sprintf(cc, "%s", "AUDIO");
		else
		    sprintf(cc, "%s", "RECEIVE");
		new_freq();		/* T & R freq may be different */
		if (hardware)
		    {
		    shift_reg = 0X0010;
		    send_shift_data(R10MHZ);	/* Hardware xmit & PA off */
		    delay(25);
		    shift_reg = 0X0000;
		    send_shift_data(R10MHZ);	/* Ant relay to receive */
		    delay(25);
	 	    rfg[rfgain].gainbyte=save_rfgain;  /* Restore rcvr gain */
		    shift_reg = 0X0004;
 /* Bob-are these needed or can new_freq() do the last  of these?? */
		    send_shift_data(R10MHZ);	/* Hardware into receive */
		    }
		error_on_tr=2;		/* Prevent error after T/R change */
		}
	setviewport(0,60,101,70,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	moveto(50,1);	outtext(cc);
	setcolor(UWHITE);

	new_freq();			/* T & R freq may be different */
	}

/* rit_toggle()  -  Turns RIT on and off.  When turned back on the old
 * offset is restored.  Not restored after shutdown.
 */
void rit_toggle(void)
	{char cc[10];

	rit_on ^= 1;			/* Toggle rit on/off */
	if (rit_on)
		{
		rit = rit_save;
		}
	else
		{
		rit_save = rit;
		rit = 0;
		}
	new_freq();
	}

/* new_band(int nband)  -  Changes parameters to nband.
 */
void new_band(int nband)
	{
	if (input_if_af)		/* Non-operative in audio mode */
	    {
	    do_error(11, 200);
	    return;
	    }
	if (tvtr[nband].lo_type==999)
	    {
	    do_error(10, 200);
	    return;			/* No transverter for this band */
	    }
	setviewport(0, 3, 102, 16, DO_CLIP);
	clearviewport();
	settextjustify(1,2);
	setcolor(URED);
	moveto(51, 3);
	outtext(tvtr[nband].name);

	tvtr[band].last_tfreq=tfreqL;	/* Save last freq and mode */
	tvtr[band].last_mode=mode;
	tfreqL=(long)tvtr[nband].last_tfreq;  /* Load new freq and mode */
	mode=tvtr[nband].last_mode;
	band=nband;
	new_freq();
	set_mode();
	}

/* The first conversion synthesizer freq is set by the long int in the
 * call to send_shift_data. The dsp lo is set by global uint's dsp_deltap_lo
 * and dsp_deltap_hi.  These are sent to the dsp by the '7' command.
 * For transmit the above frequencies are calculated here, based on
 *      global transmit		1=transmit, 0=receive
 *      global tfreqL		Transmit carrier freq in Hz at 2 meters
 *	global tvtr[band].lo_freq  Offset of displayed freq due to transverter
 *	global tvtr[band].sb	-1 if i-f=rf-lo, else +1
 *	global mode		
 * For receive, all the above are used, plus
 *	global rit_on		1 if RIT is on, else 0
 *	global rit		Offset in Hz, +RIT raises the receiver lo
 *	global cwoffset		Receive is usb and the displayed frequency
 *				  is higher by cwoffset than the zero-beat.
 * The displayed receive frequency includes RIT which is also displayed
 * separately.
 */
void new_freq(void)
	{
	int i, x;
	char cc[100];
	unsigned int cmd2;
	unsigned int dsp_dph;
	double synth_freq, dt_freq;
	long int    disp_MHz;		/* MHz part of display */
	long int    disp_Hz;
	long int    disp_f;
	long int    aL, bL, freqL;
	long int    int_freqL, synth_freqL, n_synthL;
	static long int  last_n_synthL = 0;

	/* DSP lo freq ranges from 12.501 to 17.500 kHz as main synthesizer
	   goes in 5 KHz steps.  */
	
	if (rit_on)
		setcolor(UWHITE);
	else
		setcolor(UBEIGE);
	setviewport(0,303,126,312,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(2,2);
	moveto(126,0);
	sprintf(cc, "%d", rit);
	outtext(cc);
	settextjustify(0,2);
	moveto(2,0);
	outtext("RIT");

	if (!waterfall)
	    {
	    setviewport(X2R-47, Y2T+20, X2R-1, Y2T+40, DO_CLIP);
	    clearviewport();
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(23, 1);
	    outtext("RIT");
	    moveto(23, 11);
	    if (rit_on)
		outtext(cc);
	    else
		outtext("Off");
	    }

	/* Calculate 'freqL', the carrier frequency needed at 2 meters */
	freqL = tfreqL;			/* Base for all freqs */
	if (!transmit && rit_on)
	    freqL += (long int)rit;	/* For receive only */
	if (transmit && mode==CW)
	    freqL += (long int)cwoffset;

	setviewport(2,126,125,159,DO_CLIP);
	clearviewport();
	settextjustify(1,2);
	settextstyle(0, 0, 1);
	setcolor(UWHITE);
	moveto(62,2);

	if (input_if_af==0)		/* For rf processing */
	    {
	    /* Correct for reference frequency error, first:  */
	    freqL=(long int)(0.5+ref_offset*(double)freqL);
	    int_freqL=5000L*((freqL+2500L)/5000L);
	    synth_freqL=int_freqL-19665000L;
	    n_synthL=synth_freqL/5000L;
	    dsp_dph=(unsigned int)
		(1.365333333*(double)(int_freqL-freqL+15000UL));
	    dsp_deltap_lo=0XFF & dsp_dph;
	    dsp_deltap_hi=0XFF & (dsp_dph >> 8);
	    /* In 7 cmd, byte 3 is strobe on shift regs. Set here to 0 */
	    ASYNCOUT(ESC_CHAR);      ASYNCOUT(7);   /* This sets dsp lo */
	    ASYNCOUT(0);             ASYNCOUT(1);
	    ASYNCOUT(dsp_deltap_lo);  ASYNCOUT(dsp_deltap_hi);

#if 0
 /* Temp code to readout freq setting values on waterfall area */
 setviewport(130,250,630,300,DO_CLIP);
 clearviewport();
 settextjustify(0,2);
 moveto(1, 1);
 sprintf(cc, "Freq=%ld   DSP freq=%ld", freqL, (int_freqL-freqL+15000L));
 outtext(cc);
 moveto(1,11);
 sprintf(cc,"n_synthL=%ld", n_synthL);
 outtext(cc);
#endif

	    setviewport(2,126,125,159,DO_CLIP);
	    clearviewport();
	    settextjustify(1,2);
	    settextstyle(0, 0, 1);
	    setcolor(UWHITE);
	    moveto(62,2);

	    if (n_synthL != last_n_synthL)	/* If need new synth freq */
		{
		bL=( (unsigned long)( (double)n_synthL / 64.0 ) ) & 0X7FFUL;
		aL=(n_synthL - 64UL * bL) & 0X7FUL;
		send_shift_data( (aL<<1) | (bL<<8) );  /* To dsp to prog PLL */
		last_n_synthL=n_synthL;
		}

	    /* Now find the freq to actually display.  Note that in cw the
	     * displayed freq is that which receives a signal producing
	     * a pitch of cwoffset.     */
	    dt_freq=tvtr[band].lo_freq-((double)tvtr[band].sb)*(double)tfreqL;
	    if (!transmit) dt_freq += (double)rit;
	    if (mode==CW)  dt_freq+=(double)cwoffset;
	    disp_MHz=(long int)(0.000001*(dt_freq+0.0000005));
	    disp_Hz=(long int)(dt_freq-1.0e6*(double)disp_MHz);

	    if (waterfall)
		{
		sprintf(cc, "%ld", disp_MHz);
		outtext(cc);
		settextstyle(0, 0, 2);
		if (mode==FM)
		    sprintf(cc, ".%3.3ld", disp_Hz/1000);
		else
		    sprintf(cc, ".%6.6ld", disp_Hz);
		moveto(62,15);  outtext(cc);
		settextstyle(0, 0, 1);
		}
	    else
		{
		setviewport(0,124,127,161,DO_CLIP);	/* Small Freq Box */
		clearviewport();

		/* 128, 639   d=511  */
		setviewport(X2L+53, Y2T+8, X2L+460, Y2T+39, DO_CLIP);
		clearviewport();
		for(i=0; i<5; i++)
		    {
		    x = 43+i*80;
		    line(x, 35, x, 24);
		    }
		for(i=0; i<25; i++)
		    line(11+i*16, 35, 11+i*16, 29);
		/* The frequency cursor: */
		x=3+(int)((freqL-143493750)/12500);
		setcolor(ULIGHTRED);
		line(x, 34, x, 29);
		setlinestyle(SOLID_LINE, 0, THICK_WIDTH);
		line(x, 28, x, 23);
		/* Frequency Display: */
		disp_f=1000000L*disp_MHz+disp_Hz;

		setlinestyle(SOLID_LINE, 0, NORM_WIDTH);
		setcolor(UYELLOW);
		settextjustify(CENTER_TEXT, TOP_TEXT);
		settextstyle(DEFAULT_FONT, HORIZ_DIR, 3);
		if (mode==FM)
		    {
		    if (disp_MHz < 1000)
			moveto(170, 0);
		    else if (disp_MHz>=1000 && disp_MHz<10000)
			moveto(158, 0);
		    else if (disp_MHz>=10000 && disp_MHz<100000)
			moveto(146, 0);
		    else 			/* disp_MHz>=100000 */
			moveto(134, 0);
		    sprintf(cc, "%ld.", disp_MHz);
		    outtext(cc);
		    moveto(252, 0);
		    disp_f=disp_Hz/1000;
		    sprintf(cc, "%3.3ld", disp_f);
		    outtext(cc);
		    }
		else
		    {
		    if (disp_MHz < 1000)
			moveto(143, 0);
		    else if (disp_MHz>=1000 && disp_MHz<10000)
			moveto(131, 0);
		    else if (disp_MHz>=10000 && disp_MHz<100000)
			moveto(119, 0);
		    else 			/* disp_MHz>=100000 */
			moveto(107, 0);
		    sprintf(cc, "%ld.", disp_MHz);
		    outtext(cc);
		    moveto(223, 0);
		    disp_f=disp_Hz/1000;
		    sprintf(cc, "%3.3ld", disp_f);
		    outtext(cc);
		    moveto(305, 0);
		    disp_f=disp_Hz-1000*disp_f;
		    sprintf(cc, "%3.3ld", disp_f);
		    outtext(cc);
		    }
		setcolor(UWHITE);
		}
	    }
	else
	    {				/* For audio processing */
	    dsp_dph = (unsigned int)( 1.365333333 * (double)tfreqL );
	    if (transmit)
		dsp_dph=0;
	    if (mode==CW && transmit)
		dsp_dph= (unsigned int)( 1.365333333 * (double)cwoffset );
	    dsp_deltap_lo=0XFF & dsp_dph;
	    dsp_deltap_hi=0XFF & (dsp_dph >> 8);
	    /* In 7 cmd, byte 3 is strobe on shift regs. Set here to 0 */
	    ASYNCOUT(ESC_CHAR);      ASYNCOUT(7);
	    ASYNCOUT(0);             ASYNCOUT(1);
	    ASYNCOUT(dsp_deltap_lo);  ASYNCOUT(dsp_deltap_hi);

	    disp_Hz=tfreqL;
	    outtext("Offset Hz");
	    settextstyle(0,0,2);
	    sprintf(cc, "%ld", disp_Hz);
	    moveto(62,15);  outtext(cc);
	    settextstyle(0, 0, 1);
	    }				/* End audio input */
	}

/* The af gain is set in the dsp by a multiplicative constant and a shift
 * value that ends up in the se register.  Up to 50, se=0.  Each increase
 * of 1 in se is 6 dB of added gain.
 */
void set_afgain(void)
	{char cc[10];
	int afg, ifg;
	int se;				/* Shift value for dsp se register */

	setviewport(63,393,126,402,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", afgain);
	outtext(cc);
	if(afgain==0)
	    {
	    se=0;
	    afg=0;
	    }
	else if(afgain>40)
	    {
	    se=(afgain-35)/6;
	    afg=95+(afgain-35)%6;
	    }
	else
	    {
	    se=0;
	    afg=60+afgain;
	    }
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('G');
	ASYNCOUT( fgainl[afg] ); ASYNCOUT( fgainh[afg] );
	ASYNCOUT( se ); ASYNCOUT( 4 );
	}

void set_mike_gain(void)
	{char cc[10];
	setviewport(63,258,126,267,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", mike_gain);
	outtext(cc);
	set_mode();			/* Send new mike_gain to dsp */
	}

void set_cwoffset(void)
	{char cc[30];
	int sidetone_ph;

	setviewport(34, 288, 84, 297, DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	setcolor(UBEIGE);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	moveto(25, 1);
	sprintf(cc, "%d %d", sidetone_lev, cwoffset);
	outtext(cc);
	setcolor(UWHITE);

	sidetone_ph=(unsigned int)(1.365333333*(double)cwoffset);
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('H');
	ASYNCOUT(sidetone_ph & 0XFF);  ASYNCOUT((sidetone_ph>>8) & 0XFF); 
	ASYNCOUT( fgainl[sidetone_lev] ); ASYNCOUT( fgainh[sidetone_lev] );

	new_freq();
	}

void set_fm_squelch(void)
	{char cc[10];
	int sq;

	if (mode != FM) return;
	setviewport(34, 288, 84, 297, DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(1,2);
	setcolor(UBEIGE);
	moveto(25, 1);
	sprintf(cc, "%d", fm_squelch);
	outtext(cc);
	setcolor(UWHITE);
	sq=2*(100-fm_squelch)*(100-fm_squelch);
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('S');
	ASYNCOUT(sq & 0XFF);  ASYNCOUT((sq>>8) & 0XFF); 
	ASYNCOUT(3);  ASYNCOUT(4);
	}

void set_xmit_power(void)
	{char cc[10];
	setviewport(63,278,126,287,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", xmit_power);
	outtext(cc);
	if (transmit)			/* Update dsp */
	    {ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
	     ASYNCOUT(1 | (cw_on<<4));  ASYNCOUT(2); 
	     ASYNCOUT( fgainl[xmit_power] ); ASYNCOUT( fgainh[xmit_power] );
	    }
	}

void set_rfgain(void)
	{char cc[10];
	setviewport(63,403,126,412,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", rfg[rfgain].dbdisplay);
	outtext(cc);
	send_shift_data(R10MHZ);
	}

void set_cw_speed(void)
	{char cc[10];
	setviewport(66,268,127,277,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(60, 1);
	outtext(t[cw_speed].description);
	}

void set_contrast(void)
	{char cc[10];
	setviewport(63,233,126,242,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", contrast);
	outtext(cc);
	dis_gain=0.0025+0.0035*(double)(contrast);
	contrast_line_count=20;
	}

void set_brightness(void)
	{char cc[50];
	setviewport(63,243,126,252,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", bright);
	outtext(cc);
	dis_offset=200.0-3.0*(double)bright;
	contrast_line_count=20;
	}

void set_window(void)
	{char cc[10];
	if (do_window==0)
	    {
	    sprintf(cc, "%s", "None");
	    win_adj=-2.5;
	    }
	if (do_window==1)
	    {
	    sprintf(cc, "%s", "Hamming");
	    win_adj=0.0;
	    }
	if (do_window==2)
	    {
	    sprintf(cc, "%s", "BH-92");
	    win_adj=3.55;
	    }
	setviewport(63,378,126,387,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	outtext(cc);
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('W');
	ASYNCOUT(do_window); ASYNCOUT(2);
	ASYNCOUT(3); ASYNCOUT(4);
	}

/* set_mode()  -  Arrive with global int mode=0, 1, 2,...
 */
void set_mode(void)
	{
	char cc[20];
	int cmd1, cmd2;
	/* The save_x_freq are used when going into and out of the fm mode
	   to allow fm to go in 5 kHz steps, but return to the 1 Hz freq */
	static long	save_fine_freqL=0;
	long		freq_5L;	/* Used to test if on even 5Khz freq */

	/* First clear old data from Mode line  */
	setviewport(34, 288, 126, 297, DO_CLIP);
	clearviewport();

	freq_5L=5000*(tfreqL/5000);
	if (mode==FM && freq_5L!=tfreqL)  /* Use 5 kHz resolution if not now*/
	    {
	    save_fine_freqL=tfreqL;
	    tfreqL = 5000*((tfreqL+2500)/5000);
	    }
	if (mode==(FM+1) && save_fine_freqL>0)
	    {
	    tfreqL=save_fine_freqL;	/* Get old freq to 1 Hz */
	    }
	if (mode==FM)
	    set_agc(0);
	else
	    set_agc(agc_decay);

	if (narrow)
	    {
	    switch (mode)
		{
		case CW: {
			sprintf(cc, "%s", "   CW");
			cmd1=1;		/* 1=CW */
			cmd2=0;		/* 0=USB (LSB in dsp) */
			set_cwoffset();
			print_medium(0, -1);	/* Initialize */
			break;
			}
		case USB: {
			sprintf(cc, "%s", "  USB");
			cmd1=0;		/* 0=SSB */
			cmd2=0;		/* 0=USB (LSB in dsp) */
			break;
			}
		case LSB: {
			sprintf(cc, "%s", "  LSB");
			cmd1=0;		/* 0=SSB */
			cmd2=1;		/* 1=LSB (USB in dsp) */
/* cmd2=2; 2=Binaural */
			break;
			}
		case FM: {
			sprintf(cc, "%s", "   FM");
			cmd1=2;		/* 2=FM */
			cmd2=0;		/* 0=no ssb */
			set_fm_squelch();
			break;
			}
		case 4: {
			sprintf(cc, "%s", "  LHL");
			cmd1=0x81;		/* 80=Triggered 2 sec, 1=CW */
			cmd2=0;			/* 0=USB */
			print_big(0, -1, -1);	/* Initialize */
			break;
			}
		case 5: {
			sprintf(cc, "%s", "PUA43");
			cmd1=0x81;		/* 80=Triggered 2 sec, 1=CW */
			cmd2=0;			/* 0=USB */
			print_big(0, -1, -1);	/* Initialize */
			pua43_update(-1);	/* Also initialize */
			break;
			}

		/* AM=6, ... */
		default: break;
		}
	    }
	else				/* else if wideband modes */
	    {


#if 0
This needs to be integrated into things.  The most useful is binaural
and USB left, LSB right
	    if (binaural && iq)
		{
		sprintf(cc, "%s",  "WB IQ");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X96;		/* Wide, I-Q, Binaural */
		}
	    else if (!binaural && iq)
		{
		sprintf(cc, "%s",  "WM  I");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X94;		/* Wide, I-Q, Monaural */
		}
	    else if (binaural && !iq)
		{
		sprintf(cc, "%s",  "B SSB");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X9A;		/* Wide, SSB, Binaural */
		}
	    else if (!binaural && !iq)
		{
		sprintf(cc, "%s",  "M USB");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X98;		/* Wide, SSB, Monaural */
		}
#endif

	    }
	new_freq();			/* Get cw offset, etc */
	setviewport(84,288,126,297,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(42, 1);
	outtext(cc);

	if (narrow)  sprintf(cc, "%s",  "     "); /* Temporary solution */
	if (!waterfall)
	    {
	    setviewport(X2L+1, Y2T+25, X2L+49, Y2T+35, DO_CLIP);
	    clearviewport();
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(24, 1);
	    outtext(cc+2);
	    }
	/* Note: when these commands get to dsp they produce the reverse
	   sideband.  This is because of upper side conversion at 19.68 MHz.
	   Notation, USB, LSB here refers to signal at 2 meters */
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('M');
	ASYNCOUT(cmd1); ASYNCOUT(cmd2);
	ASYNCOUT( fgainl[mike_gain] ); ASYNCOUT( fgainh[mike_gain] );
	}

void set_data_save(void)
	{
	do_data_save ^= 1;	
	setviewport(0,80,101,90,DO_CLIP);
	clearviewport();
	setcolor(UBEIGE);
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	moveto(50,1);
	if (do_data_save)
	    outtext("RECORDING");
	}

void normalize_plot(int k)
	{
	int i;

	for (i=0; i<512; i++)
		dnorm[i]=0.0;
	setviewport(0,80,101,90,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	moveto(50,1);
	if (k==0)			/* Clear normalization */
	    {
	    normalize_on=0;
	    }
	else if (k==1)			/* Add fast normalization */
	    {
	    normalize_on=25;		/* Count down to 1 */
	    normalize_num=25.0;
	    outtext("Wait...");
	    }
	else
	    {
	    normalize_on=1000;		/* Big countdown for precision */
	    normalize_num=1000.0;
	    outtext("Wait...");
	    }
	}

/* do_auto_disp(k)  -  corrects for analog gain variations so that the
 * colors in the spectral display don't drift.  The down side to this
 * is that strong signals tend to push the noise down.  So it works very
 * well for weak signals in noise.
 */
void do_auto_disp(int k)
	{
	int i;

	setviewport(63,423,126,432,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	if (k==0)			/* Stop color agc */
	    {
	    auto_disp_on=FALSE;
	    setcolor(UBEIGE);
	    outtext("OFF");
	    }
	else				/* Turn on auto_disp */
	    {
	    auto_disp_on=-1;		/* A flag to initialize base value */
	    setcolor(UWHITE);
	    outtext("ON");
	    }
	}

/* set_8_16(k)  -  Programs the DSP for 8 or 16 bit data (either 512 bytes
 * or 256 words of logged spectral data).  Changes the resolution of the
 * data coming to the PC from the DSP.  If k=0, resets bits_8_16=0, the
 * value for 8 bit data.  If k=1 it toggles 8/16.  If k=9 it just programs
 * the DSP to the current value of bits_8_16.
 */
void set_8_16(int k)
	{
	if (k==0)  bits_8_16=0;		/* Reset to 8 bit on 0 */
	else if (k==1)  bits_8_16 ^= 1;		/*   or toggle on 1 */
	/* Fall through on 9 */
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('D');
	ASYNCOUT(bits_8_16);
	ASYNCOUT(2); ASYNCOUT(3); ASYNCOUT(4);
	}

/* set_input()  -  Arrive with global input_if_af=0 for i-f or 1 for af.
 */
void set_input(void)
	{
	static long	save_af_freqL=0;
	static long	save_tfreqL=146000000;
	char cc[50];
	
	if (transmit) t_r_toggle();	/* Either if or af, it's receive */

	setviewport(0,60,101,70,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	setcolor(UWHITE);
	moveto(50,1);
	if (input_if_af==0)
		{
		outtext("RECEIVE");
		save_af_freqL=tfreqL;	/* Going from af to if input */
		tfreqL=save_tfreqL;
		set_mode();
		}
	else
		{
		outtext("AUDIO");
		save_tfreqL=tfreqL;	/* Going from if to af input */
		tfreqL=save_af_freqL;
		if (mode==USB)
		    {
		    mode=LSB;
		    set_mode();
		    }
		ASYNCOUT(ESC_CHAR);  ASYNCOUT('M');
		ASYNCOUT(0); ASYNCOUT(1);	/* This sets LSB (dsp usb)*/
		ASYNCOUT( fgainl[mike_gain] ); ASYNCOUT( fgainh[mike_gain] );
		}
	new_freq();

	setviewport(0, 3, 102, 16, DO_CLIP); /* Now transverter designation */
	clearviewport();
	settextjustify(1,2);
	setcolor(URED);
	moveto(51, 3);
	if (input_if_af==0)  outtext(tvtr[band].name);

	setcolor(UWHITE);
	setviewport(2,101,104,111,DO_CLIP);
	clearviewport();
	if (input_if_af)		/* Only for audio processor */
	    {
	    sprintf(cc, "Fd=%dMHz", near_int(0.000001*f_eme) );
	    settextstyle(0,0,1);
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(51, 0);
	    outtext(cc);
	    }

	ASYNCOUT(ESC_CHAR);  ASYNCOUT('J');
	ASYNCOUT(input_if_af);
	ASYNCOUT(0); ASYNCOUT(3); ASYNCOUT(4);   /* Audio to L & R */
	}

/* set_affilt_sw()  -  Arrive with do_af_filt=0 for no af filter.
 */
void set_affilt_sw(void)
	{
	annotate_af_filt();
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('A');
	ASYNCOUT(do_af_filt);
	ASYNCOUT(do_denoise);
	ASYNCOUT(3); ASYNCOUT(4);
	}

/* set_affilt_select()  -  Arrive with af_coef 0 or 1.  The dsp command
 * selects filter 0 or 1 with the lsb and sets the gain with the higher
 * bits.  Gain of 1, 8, or 64.
 */
void set_affilt_select(void)
	{
	annotate_af_filt();
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('B');
	ASYNCOUT(fir3[af_coef]);
	ASYNCOUT(2); ASYNCOUT(3); ASYNCOUT(4);
	}

/* set_delay_right(do_what)   Sets amount of delay put into right channel
 * in dsp.  do_what=0 means toggle delay on and off.
 */
void set_delay_right(int do_what)
	{

	int del;
	char cc[20];

	del=(int)(0.001+pow(2.0, (double)(delay_right+1)));
	if (do_what==0)
	    {
	    delay_state ^= 1;
	    }
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('E');
	setviewport(63,333,126,342,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	if (delay_state)
	    {
	    ASYNCOUT(del & 0XFF); ASYNCOUT((del>>8) & 0XFF);
	    sprintf(cc, "%d", del);
	    outtext(cc);
	    }
	else
	    {
	    outtext("Off");
	    ASYNCOUT(1);  ASYNCOUT(0);	/* Delay off is delay of 1 */
	    }
	ASYNCOUT(3); ASYNCOUT(4);
	}

void select_denoise(void)
	{
	do_denoise ^= 1;
	set_denoise();			/* This updates the screen */
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('A');
	ASYNCOUT(do_af_filt);
	ASYNCOUT(do_denoise);
	ASYNCOUT(3); ASYNCOUT(4);
	}

void set_denoise(void)
	{char cc[10];
	int denoise_beta;

	/* As denoise_level ranges (0, 100) denoise_beta ranges (30000, 300) */
	denoise_beta =(int)(30000.0/pow(1.04712, (double)denoise_level));

	setviewport(63,323,126,332,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	if (do_denoise)
	    {
	    sprintf(cc, "%d", denoise_level);
	    outtext(cc);
	    }
	else
	    outtext("Off");
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('C');
	ASYNCOUT(denoise_beta & 0XFF); ASYNCOUT((denoise_beta>>8) & 0XFF);
	ASYNCOUT(3); ASYNCOUT(4);
	}

void set_agc(int decay)
	{char cc[20];
	static int agcd[]={0, 32, 24, 16, 12, 8, 6, 4, 3, 2, 1};

	setviewport(63,413,126,422,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(64, 1);
	if (decay>0)
	    {
	    sprintf(cc, "%d", decay);
	    outtext(cc);
	    }
	else
	    outtext("Off");
	/* AGC Settings: */
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('V');
	ASYNCOUT(2); ASYNCOUT(3);
	ASYNCOUT(agcd[decay]&0XFF); ASYNCOUT((agcd[decay]>>8)&0XFF);
	}

void annotate_af_filt(void)
	{
	char cc[10];
	if (do_af_filt==0)   sprintf(cc, "%s", "None");
	else if (af_coef==0) sprintf(cc, "%s", "CW450");
	else                 sprintf(cc, "%s", "CW200");
	setcolor(UWHITE);
	setviewport(63,313,126,322,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	outtext(cc);
	}

void set_dot(void)
	{dot_type += 1;
	 if  (dot_type>2) dot_type=0;
	}

/* disp_toggle()  As a resuly of the Alt Y command sets the display format
 * to either Large freq display, or to waterfall.
 */
void disp_toggle(void)
	{
	char cc[20];

	setviewport(0, 232, 70, 252, DO_CLIP);
	clearviewport();
	settextjustify(0,2);
	settextstyle(0,0,1);
	if (waterfall)
	    {
	    setcolor(UWHITE);
	    moveto(2,1);	outtext("Contrast");
	    moveto(2,11);	outtext("Bright");
	    setviewport(0, 0, 639, 479, DO_CLIP);
	    line(X2L, Y2T, X2R, Y2T);
	    setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
	    clearviewport();
	    scroll_line=0;
	    setcolor(URED);			/* Freq display border */
	    setviewport(0, 124, 127, 161, DO_CLIP);
	    rectangle(0, 0, 127, 37);
	    rectangle(1, 1, 126, 36);
	    setcolor(UWHITE);
	    new_freq();
	    }
	else
	    {
	    setcolor(UBEIGE);
	    moveto(2,1);	outtext("Contrast");
	    moveto(2,11);	outtext("Bright");
	    setcolor(UWHITE);
	    draw_panel();
	    new_freq();
	    switch (mode)
		{
		case CW: {
			sprintf(cc, "%s", "   CW");
			break;
			}
		case USB: {
			sprintf(cc, "%s", "  USB");
			break;
			}
		case LSB: {
			sprintf(cc, "%s", "  LSB");
			break;
			}
		case FM: {
			sprintf(cc, "%s", "   FM");
			break;
			}
		default: break;
		}
	    setviewport(X2L+1, Y2T+15, X2L+49, Y2T+45, DO_CLIP);
	    clearviewport();
	    settextstyle(0, 0, 1);
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(24, 10);
	    outtext(cc+2);
	    }
	if (mode==CW)
	    {
	    morse(ADD_CH, 32);		/* Force a redraw of cw text */
	    morse(ADD_CH, 8);		/* Back it out */
	    }
	}
	    
/* set_display_width()  -  Arrive with global display_width 0 or 1,
 * i.e., FULL_FREQ or HALF_FREQ.
 */
void set_display_width(void)
	{
	if (display_width)
		mark_frequency(2000);
	else
		mark_frequency(4000);
	}

/* set_db_div(changed) is used to change the setting (changed==1) and
 * to redraw the display (changed==0).
 */
void set_db_div(int changed)
	{char cc[10];
	int i, yy, kdb;

	if (db_div_set==0)
		{
		kdb_div=3.2;
		kdb=10;
		}
	else if (db_div_set==1)
		{
		kdb_div=6.4;
		kdb=5;
		}
	else if (db_div_set==2)
		{
		kdb_div=16.0;
		kdb=2;
		}
	else if (db_div_set==3)
		{
		kdb_div=32.0;
		kdb=1;
		}
	setviewport(X1L+1, Y1T+2, X1L+68, Y1T+11, DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(0,2);
	setcolor(UBEIGE);
	moveto(1, 1);
	sprintf(cc, "%d dB/", kdb);
	outtext(cc);

	if (changed)
	    {
	    setviewport(102, Y1T, X1L-1, Y1B-8, DO_CLIP);
	    clearviewport();
	    settextjustify(1,1);
	    for (i=0; i<=7; i++)
		{
		yy =  1 + i*16;
		moveto(14, yy);
		sprintf(cc, "%d", kdb*(8 -i));
		outtext(cc);		/* dB level markings */
		}
	    setcolor(UWHITE);
	    set_disp_offset(1);		/* Check DSP scaling is OK */
	    }
	}

void set_disp_offset(int update_dsp)
	{
	char	cc[30];
	if ( near_int(koffset) != 0)
		{
		setviewport(X1R-90, Y1T+1, X1R-4, Y1T+11, DO_CLIP);
		clearviewport();
		settextstyle(0,0,1);
		settextjustify(2,2);
		setcolor(UBEIGE);
		moveto(84, 1);
		sprintf(cc, "%d dB", near_int(koffset) );
		outtext(cc);
		setcolor(UWHITE);
		}
	/* power_exp is sent to the dsp to scale the power data before it is
	   log converted to 8 bits.  The total dynamic range of the 8 bit data
	   is 80 dB (lsb=80/256) and this factor keeps it centered in the
	   80 dB display area.
	   In the following we add 99.0 to the offset to prevent (int)
	   confusion at 0.  It is removed by an extra 33 in front.
	   Note: not used except with 10 dB/div.  Therefore you can still
	   overload if you offset more than -50 dB in 5 dB/div.
	   Limit to this amount back in command.			*/
	if (db_div_set==0)		/* 0 --> 10 dB/div  */
	    power_exp=33+10-((int)(99.0+koffset))/3;
	else
	    power_exp=10;

	if (update_dsp)
	    {
	    ASYNCOUT(ESC_CHAR);  ASYNCOUT('D');
	    ASYNCOUT(bits_8_16);
	    ASYNCOUT(power_exp);
	    ASYNCOUT(LOG_OFFSET&0XFF); ASYNCOUT((LOG_OFFSET>>8)&0XFF);
	    }
	}

void set_noncohave(void)
	{char cc[50];
	sprintf(cc, "%d", noncohave);
	setviewport(63,368,126,377,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);	outtext(cc);
	}

/* read_config_file()  -  Restores transceivers to previous settings
 * using UHFA.CFG.
 * Returns:	 0  Non-fatal error, or no file found.
 *		 1  Successful file read.
 */
int read_config_file(void)
	{
	int i;
	static char entrysep[]=" \t,=\n\r";	/* Allow sp, tab, comma,
						   equal, lf, cr  */
	char *ci, *cd;
	char cline[260], cline2[260];

        if (NULL == ( config_file=fopen("UHFA.CFG", "rt") ))
	    {
	    printf("No UHFA.CFG file.  Using default configuration\n");
	    return 0;
	    }

	if (fgets(cline, 256, config_file) == NULL)
	    {
	    printf("Empty UHFA.CFG file.  Using default configuration\n");
	    fclose(config_file);
	    return 0;
	    }
 	ci = strtok(cline, entrysep);	/* Should be file identifier line */
	if (strcmp(ci, "UHFACFG2"))
	    {
	    printf("Illegal config file, not UHFACFG2. Use default config\n");
	    fclose(config_file);
	    return 0;
	    }
	while (fgets(cline, 256, config_file))
	    {
	    strcpy(cline2, cline);	/* Extra for when string token first */
	    /* Get identifier and skip return:   */
	    if ((ci = strtok(cline, entrysep)) == NULL)  break;
	    /* Get a token to cd */
	    if ((cd = strtok(NULL, entrysep)) == NULL)  strcpy(cd, "BLANK");

	    if      (!stricmp(ci, "contrast")) contrast=atoi(cd);
	    else if (!stricmp(ci, "bright"))  bright=atoi(cd);
	    else if (!stricmp(ci, "afgain"))  afgain=atoi(cd);
	    else if (!stricmp(ci, "rfgain"))  rfgain=atoi(cd);
	    else if (!stricmp(ci, "rfgdb"))
		{
		rfg[1].dbgain=atof(cd);
		for (i=2; i<=6; i++)
		    {
		    if ((cd=strtok(NULL, entrysep))==0)
			{
			printf("Missing rfgdb entry in UHFA.CFG (need 6)\n");
			return 0;
			}
		    else
			{
			rfg[i].dbgain=atof(cd);
			}
		    }
		}
	    else if (!stricmp(ci, "gain_adj"))  gain_adj=atof(cd);
	    else if (!stricmp(ci, "cw_speed"))  cw_speed=atoi(cd);
	    else if (!stricmp(ci, "agc_decay"))  agc_decay=atoi(cd);
	    else if (!stricmp(ci, "mike_gain"))  mike_gain=atoi(cd);
	    else if (!stricmp(ci, "xmit_power"))  xmit_power=atoi(cd);
	    else if (!stricmp(ci, "fm_squelch"))  fm_squelch=atoi(cd);
	    else if (!stricmp(ci, "cwoffset"))  cwoffset=atoi(cd);
	    else if (!stricmp(ci, "sidetone_lev"))  sidetone_lev=atoi(cd);
	    else if (!stricmp(ci, "noncohave"))  noncohave=atoi(cd);
	    else if (!stricmp(ci, "do_window"))  do_window=atoi(cd);
	    else if (!stricmp(ci, "do_af_filt"))  do_af_filt=atoi(cd);
	    else if (!stricmp(ci, "af_coef"))  af_coef=atoi(cd);
	    else if (!stricmp(ci, "do_denoise"))  do_denoise=atoi(cd);
	    else if (!stricmp(ci, "denoise_level"))  denoise_level=atoi(cd);
	    else if (!stricmp(ci, "dot_type"))  dot_type=atoi(cd);
	    else if (!stricmp(ci, "display_width"))  display_width=atoi(cd);
	    else if (!stricmp(ci, "bin_low"))  bin_low=atoi(cd);
	    else if (!stricmp(ci, "db_div_set"))  db_div_set=atoi(cd);
	    else if (!stricmp(ci, "koffset"))  koffset=atof(cd);
	    else if (!stricmp(ci, "mode"))  mode=atoi(cd);
	    else if (!stricmp(ci, "bits_8_16"))  bits_8_16=atoi(cd);
	    else if (!stricmp(ci, "input_if_af"))  input_if_af=atoi(cd);
	    else if (!stricmp(ci, "hardware"))  hardware=atoi(cd);
	    else if (!stricmp(ci, "rit"))  rit=atoi(cd);
	    else if (!stricmp(ci, "rit_on"))  rit_on=atoi(cd);
	    else if (!stricmp(ci, "band"))  band=atoi(cd);
	    else if (!stricmp(ci, "tfreqL"))  tfreqL=atol(cd);
	    else if (!stricmp(ci, "ref_offset"))  ref_offset=atof(cd);
	    else if (!stricmp(ci, "diag_mode"))  diag_mode=atoi(cd);
	    else if (!stricmp(ci, "startup_detail")) startup_detail=atoi(cd);
	    else if (!stricmp(ci, "waterfall"))  waterfall=atoi(cd);
	    else if (!stricmp(ci, "shift_reg"))  shift_reg=atoi(cd);
	    else if (!stricmp(ci, "time_type"))  time_type=atoi(cd);
	    else if (!stricmp(ci, "time_str"))
		{
		cd=strtok(cline2, "|");		/* Go to time_str */
		cd=strtok(NULL, "|");		/* Get string */
		strcpy(time_str, cd);
		}
	    else if (!stricmp(ci, "call_str"))
		{
		cd=strtok(cline2, "|");		/* Go to call_str string */
		cd=strtok(NULL, "|");		/* Get string */
		strcpy(call_str, cd);
		}
	    else if (!stricmp(ci, "quit_save_state")) quit_save_state=atoi(cd);
	    else if (!stricmp(ci, "beep_on_error"))  beep_on_error=atoi(cd);
	    else if (!stricmp(ci, "palette"))
		{
		upalette.colors[0]=atoi(cd);
		for (i=1; i<16; i++)
		    {
		    if ((cd=strtok(NULL, entrysep))==0)
			{
			printf("Missing color entry in UHFA.CFG\n");
			return 0;
			}
		    else
			{
			upalette.colors[i]=atoi(cd);
			}
		    }
		}
	    else if ( !stricmp(ci,"tvtr") && (i=atoi(cd))>=0 && i<=20 )
		{
		cd=strtok(NULL, "|");		/* Move to string */
		cd=strtok(NULL, "|");		/* Get name string */
		strcpy(tvtr[i].name, cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].lo_type=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].sb=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].lo_freq=atof(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].chain_mult=atoi(cd);
		strtok(NULL, "|");
		cd=strtok(NULL, "|");
		strcpy(tvtr[i].init_str, cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].gain_db=atof(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].control=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].bus_adr=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].last_tfreq=atof(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].last_mode=atoi(cd);
		}
	    else if ( !stricmp(ci,"loc") && (i=atoi(cd))>=0 && i<=9 )
		{
		loc[i].time_zone=0;		/* Mark as existant */
		cd=strtok(NULL, "|");		/* Move to string */
		cd=strtok(NULL, "|");		/* Get short name string */
		if (strlen(cd)>6)
		    {
		    printf("\nLoc[%d].nm exceeds 6 ch; Loc not used", i);
		    loc[i].time_zone=999;
		    }
		else
		    strcpy(loc[i].nm, cd);
		strtok(NULL, "|");
		cd=strtok(NULL, "|");		/* Get description */
		if (strlen(cd)>31)
		    {
		    printf("\nLoc[%d].name exceeds 31 ch; Loc not used", i);
		    loc[i].time_zone=999;
		    }
		else
		    strcpy(loc[i].name, cd);
		cd=strtok(NULL, entrysep);
		loc[i].lat=d2r(atof(cd));
		cd=strtok(NULL, entrysep);
		loc[i].lon=d2r(atof(cd));
		}
	    else if (!stricmp(ci, "f_eme"))  f_eme=atof(cd);
	    else if (!stricmp(ci, "cur_loc"))  cur_loc=atoi(cd);
	    else if (!stricmp(ci, "delay_state"))  delay_state=atoi(cd);
	    else if (!stricmp(ci, "delay_right"))  delay_right=atoi(cd);
	    /* Note: comments will fall through here */
	    }
	fclose(config_file);
	return 1;
	}

/* save_state()  -  Used at shut down to bring the state of the transceiver
 * back to last state.  This will replace uhfa.cfg file.
 * The first line must be "UHFACFG2".  The fprintf below summarizes the
 * the flexible format.  The special rules for strings come from strtok()
 * that gags on empty strings.  The use of the '|' allows spaces within the
 * string for titles like "432 MHz Tvtr".
 */
void save_state(void)
	{
	int i;

	if (NULL==( config_file=fopen("UHFA.CFG", "wt") ))
	    do_error(80, 200);
	else
	    {
	    rewind(config_file);
	    fprintf(config_file, "UHFACFG2\n");
	    fprintf(config_file,
		"// Note: Use double space around .name and .init_str\n");
	    fprintf(config_file,
		"// and demark each string with vertical bars '|'.\n");
	    fprintf(config_file,
		"// Empty strings should have at least a space in them.\n");
	    fprintf(config_file,
		"// Comments start with // and go to end of line\n");
	    fprintf(config_file,
		"// Fields can be delimited by space, tab, ',' ';' or '='\n");
	    fprintf(config_file,
		"//  \n");
	    fprintf(config_file,
		"// Up  to 10 locations are given by:\n");
	    fprintf(config_file,
		"//     'loc i  |Name|  |Description|  lat lon'\n");
	    fprintf(config_file,
		"// where i is (0,9) and lat and lon are decimal degrees,\n");
	    fprintf(config_file,
		"// with West being negative longitude. i=0 is home loc.\n");
	    for (i=0; i<=9; i++)
	      {
	      if (loc[i].time_zone != 999)
		    {
		    if (strlen(loc[i].nm)==0)
			strcpy(loc[i].nm, " ");
		    if (strlen(loc[i].name)==0)
			strcpy(loc[i].name, " ");
		    fprintf(config_file,
			"loc %d  |%s|  |%s|  %.5lf %.5lf\n",
			i, loc[i].nm, loc[i].name,
			r2d(loc[i].lat), r2d(loc[i].lon));
		    }
	      }
	    fprintf(config_file, "cur_loc %d\n", cur_loc);
	    fprintf(config_file,
		"// f_eme in Hz, used for audio processing only\n");
	    fprintf(config_file, "f_eme %.1lf\n", f_eme);
	    for (i=0; i<21; i++)
	      {
	      if (tvtr[i].lo_type < 999)
		{
		if (strlen(tvtr[i].name)==0)
			strcpy(tvtr[i].name, " ");
		if (strlen(tvtr[i].init_str)==0)
			strcpy(tvtr[i].init_str, " ");
	        fprintf(config_file,
		    "tvtr %d  |%s|  %d %d %.1f %d  |%s|  %.3f %d %d %.1f %d\n",
		    i, tvtr[i].name, tvtr[i].lo_type, tvtr[i].sb,
		    tvtr[i].lo_freq, tvtr[i].chain_mult, tvtr[i].init_str,
		    tvtr[i].gain_db, tvtr[i].control, tvtr[i].bus_adr,
		    tvtr[i].last_tfreq, tvtr[i].last_mode);
		}
	      }
	    fprintf(config_file, "contrast %d\n", contrast);
	    fprintf(config_file, "bright %d\n",  bright);
	    fprintf(config_file, "afgain %d\n",  afgain);
	    fprintf(config_file, "rfgain %d\n",  rfgain);
	    fprintf(config_file,
		"// The following six entries are the actual attenuation\n");
	    fprintf(config_file,
		"// for rf gain settings 94, 88,...,64. Nominal values\n");
	    fprintf(config_file,
		"// of -6.0, -12.0,...,-36.0.\n");
	    fprintf(config_file,
		"rfgdb %.2lf %.2lf %.2lf %.2lf %.2lf %.2lf\n", 
		rfg[1].dbgain, rfg[2].dbgain, rfg[3].dbgain, 
		rfg[4].dbgain, rfg[5].dbgain, rfg[6].dbgain);
	    fprintf(config_file,
		"// gain_adj is small dB value to adjust power displays.\n");
	    fprintf(config_file, "gain_adj %.2lf\n", gain_adj);
	    fprintf(config_file, "cw_speed %d\n",  cw_speed);
	    fprintf(config_file, "agc_decay %d\n",  agc_decay);
	    fprintf(config_file, "mike_gain %d\n",  mike_gain);
	    fprintf(config_file, "xmit_power %d\n",  xmit_power);
	    fprintf(config_file, "fm_squelch %d\n",  fm_squelch);
	    fprintf(config_file, "cwoffset %d\n",  cwoffset);
	    fprintf(config_file, "sidetone_lev %d\n",  sidetone_lev);
	    fprintf(config_file, "noncohave %d\n",  noncohave);
	    fprintf(config_file,
		"// do_window=1, 2 or 3 for None, Hamming or BH-92. Can\n");
	    fprintf(config_file,
		"// set here=0 for diagnostic, but will not rotate through\n");
	    fprintf(config_file,
		"// 0 with Alt W command, only 1 and 2.\n");
	    fprintf(config_file, "do_window %d\n",  do_window);
	    fprintf(config_file, "do_af_filt %d\n",  do_af_filt);
	    fprintf(config_file, "af_coef %d\n",  af_coef);
	    fprintf(config_file, "do_denoise %d\n",  do_denoise);
	    fprintf(config_file, "denoise_level %d\n",  denoise_level);
	    fprintf(config_file, "delay_state %d\n", delay_state);
	    fprintf(config_file, "delay_right %d\n", delay_right);
	    fprintf(config_file, "dot_type %d\n",  dot_type);
	    fprintf(config_file, "display_width %d\n",  display_width);
	    fprintf(config_file, "// bin_low*9.375 is first freq marked\n");
	    fprintf(config_file, "bin_low %d\n",  bin_low);
	    fprintf(config_file, "db_div_set %d\n",  db_div_set);
	    fprintf(config_file, "koffset %lf\n",  koffset);
	    fprintf(config_file, "mode %d\n",  mode);
	    fprintf(config_file, "bits_8_16 %d\n",  bits_8_16);
	    fprintf(config_file, "input_if_af %d\n",  input_if_af);
	    fprintf(config_file, "hardware %d\n",  hardware);
	    fprintf(config_file, "rit %d\n",  rit);
	    fprintf(config_file, "rit_on %d\n",  rit_on);
	    fprintf(config_file, "band %d\n",  band);
	    if (input_if_af==0)
		fprintf(config_file, "tfreqL %ld\n",  tfreqL);
	    fprintf(config_file, "ref_offset %15.13lf\n",  ref_offset);
	    fprintf(config_file, "diag_mode %d\n",  diag_mode);
	    fprintf(config_file,
		"// startup_detail=1 for diagnostic text screens:\n");
	    fprintf(config_file, "startup_detail %d\n",  startup_detail);
	    fprintf(config_file,
		"// waterfall=1 for weak-signal waterfall,\n");
	    fprintf(config_file,
		"//          =0 for big dial & help summary screen\n");
	    fprintf(config_file, "waterfall %d\n",  waterfall);
	    fprintf(config_file, "shift_reg %d\n",  shift_reg);
	    fprintf(config_file, "quit_save_state %d\n",  quit_save_state);
	    fprintf(config_file,
		"palette %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n", 
		upalette.colors[0],  upalette.colors[1],  upalette.colors[2],
		upalette.colors[3],  upalette.colors[4],  upalette.colors[5],
		upalette.colors[6],  upalette.colors[7],  upalette.colors[8],
		upalette.colors[9],  upalette.colors[10], upalette.colors[11],
		upalette.colors[12], upalette.colors[13], upalette.colors[14],
		upalette.colors[15]);
	    fprintf(config_file, "beep_on_error %d\n", beep_on_error);
	    fprintf(config_file,
		"// time_type=0 is GMT display, 1=computer time\n");
	    fprintf(config_file, "time_type %d\n", time_type);
	    if (strlen(time_str)==0)
		strcpy(time_str, ".");
	    fprintf(config_file, "time_str  |%s|\n", time_str);
	    fprintf(config_file,
		"// call_str is a string printed at top of upper-left\n");
	    fprintf(config_file,
		"// identification box, e.g. |W7XYZ|\n");
	    if (strlen(call_str)==0)
		strcpy(call_str, ".");
	    fprintf(config_file, "call_str  |%s|\n", call_str);
	    if (fclose(config_file) != NULL)
		do_error(81, 200);
	    }
	}

/* do_gif_file()  -  Records entire graphics screen in .GIF file.
 * The .GIF file has a name 'Umddhhmm.GIF' where m is the month (1,C),
 * hhmm is the hour and minute.
 * The screen must be in VGA 640x480, 16 colors.
 * Based on S. H. Huseby (gifsave.c)
 */
void do_gif_file(void)
	{
	#define NUMCOLORS         16  /* Number of different colors */
	#define BITS_PR_PRIM_COLOR 2  /* Two bits per primary color */

	int q,                      /* Counter */
	    color,                  /* Temporary color value */
	    red[NUMCOLORS],         /* Red component for each color */
	    green[NUMCOLORS],       /* Green component for each color */
	    blue[NUMCOLORS];        /* Blue component for each color */
	struct palettetype pal;
	int ic;
	char cc[15], ct[15];
	time_t t0;
	struct tm *tgp;

	if (!screen_save) return;

	setviewport(0,90,101,100,DO_CLIP);
	clearviewport();
	setcolor(ULIGHTRED);
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	moveto(50,1);
	outtext("*");      /*  GIF SAVE");  */

	cc[0]='U';
	time(&t0);
	tgp=gmtime(&t0);
	itoa( (1+tgp->tm_mon), ct, 13);   /* Jan=1,..., Dec=C */
	cc[1]=ct[0];
	itoa(tgp->tm_mday, ct, 10);
	if (tgp->tm_mday < 10)
	    {
	    cc[2]='0';
	    cc[3]=ct[0];
	    }
	else
	    {
	    cc[2]=ct[0];
	    cc[3]=ct[1];
	    }
	itoa( (100*tgp->tm_hour + tgp->tm_min), ct, 10);
	if (tgp->tm_hour < 10)
	    {
	    cc[4]='0'; cc[5]=ct[0]; cc[6]=ct[1]; cc[7]=ct[2];
	    }
	else
	    {
	    cc[4]=ct[0]; cc[5]=ct[1]; cc[6]=ct[2]; cc[7]=ct[3];
	    }
	cc[8]='.'; cc[9]='G'; cc[10]='I'; cc[11]='F'; cc[12]=0;

	setviewport(0, 0, 639, 479, 1);	/* Needed to start gif right */

	/* Get the color palette, and extract the red, green and blue
 	*  components for each color. In the EGA/VGA palette, colors are
	*  stored as bits in bytes:
	*      00rgbRGB
	*  where r is low intensity red, R is high intensity red, etc.
	*  We shift the bits in place like
	*      000000Rr
	*  for each component					*/
	getpalette(&pal);
	for (q = 0; q < NUMCOLORS; q++)
	    {
            color = pal.colors[q];
            red[q]   = ((color & 4) >> 1) | ((color & 32) >> 5);
            green[q] = ((color & 2) >> 0) | ((color & 16) >> 4);
            blue[q]  = ((color & 1) << 1) | ((color & 8) >> 3);
	    }

	GIF_Create(cc, 640, 480, NUMCOLORS, BITS_PR_PRIM_COLOR);

	/*  Set each color according to the values extracted from
	 *  the palette    */
	for (q = 0; q < NUMCOLORS; q++)
	GIF_SetColor(q, red[q], green[q], blue[q]);

	/*  Store the entire screen as an image using the user defined
	 *  callback function gpixel() to get pixel values from the screen */
	GIF_CompressImage(0, 0, -1, -1, gpixel);

	GIF_Close();

	setviewport(0,90,101,100,DO_CLIP);
	clearviewport();
	}

/* warm_boot()  -  This is a full reset of the DSP.
 */
void do_warm_boot(void)
	{
	int i;
	long a, b;

	for (i=0; i<12; i++)		/* Flush buffers, uarts etc. */
	    ASYNCOUT(0);

	ASYNCOUT(ESC_CHAR);  ASYNCOUT(0);	/* Clear dsp cmd buff */
	ASYNCOUT(0); ASYNCOUT(0);
	ASYNCOUT(0); ASYNCOUT(0);

	/* Now reprogram dsp to current states */

	ASYNCOUT(ESC_CHAR);  ASYNCOUT('J');
	ASYNCOUT(input_if_af);
	ASYNCOUT(0); ASYNCOUT(3); ASYNCOUT(4);   /* Audio to L & R channels */

	set_mode();
	set_afgain();
	set_rfgain();
	set_agc(agc_decay);
	set_window();
	set_8_16(9);
	set_affilt_sw();
	set_affilt_select();
	set_denoise();
	set_delay_right(delay_right);
	new_freq();
	set_display_width();
	}

/*
 * draw_grid()  - Assumes that BGI graphics are initialized
 * and draws grid for plotting of SA levels and threshold crossings.
 * Adds annotation for readability.
 * Set up for VGA 640 x 480 pixels ONLY
 */
void draw_grid(void)
	{
	unsigned int x, i;
	int xx, yy;
	char cc[50];

	new_freq();

	setviewport(0,0,639,479,DO_CLIP);
	setcolor(URED);
	line(3,19,101,19); line(3,53,101,53);
	line(3,19,3,53); line(101,19,101,53);
	setcolor(UYELLOW);
	settextstyle(0,0,1);
	settextjustify(1,2);
	moveto(51,23); outtext(call_str);
	moveto(51,32); outtext("DSP TRANSCVR");
	moveto(51,43); outtext("W7PUA DSP10");

	setcolor(URED);				/* Freq display border */
	rectangle(0,125,127,161);
	rectangle(1,124,126,160);

	setcolor(UWHITE);
	moveto(51,61);	outtext("RECEIVE");
	if (!hardware)
	    {
	    moveto(51, 71);
	    outtext("No RF Hdw");
	    }

	if (input_if_af)		/* Only for audio processor */
	    {
	    moveto(51, 101);
	    sprintf(cc, "Fd=%dMHz", near_int(0.000001*f_eme) );
	    outtext(cc);
	    }
	for(i=1; i<=9; i++)
		{
		x = X2L + (int)(53.229 * (double)i);
		line(x, Y2T-2, x, Y2T);		/* Vertical ticks */
		}

	line(X2L, Y2T, X2R, Y2T);		/* Lower spectral box */
	line(X2L, Y2B, X2R, Y2B);
	line(X2L, Y2T, X2L, Y2B);
	line(X2R, Y2T, X2R, Y2B);

	/* Now status info along left side */
	settextjustify(1,2);
	settextstyle(0, 0, 1);
	setcolor(UBEIGE);
	moveto(64,S_METER_POS-5);	outtext("Signal Level");

	settextjustify(0,2);
	settextstyle(0,0,1);
	setcolor(UBEIGE);
	line(1,229,127,229);
	setcolor(UWHITE);
	moveto(2,233);	outtext("Contrast");
	moveto(2,243);	outtext("Bright");
	setcolor(UBEIGE);
	line(1,255,127,255);
	setcolor(UWHITE);
	moveto(2,258);  outtext("MikeGain");
	moveto(2,268);  outtext("CW Speed");
	moveto(2,278);  outtext("Xmit Pwr");
	moveto(2,288);	outtext("Mode");
	setcolor(UBEIGE);
	line(1,299,127,299);
	moveto(2,303);	outtext("RIT");
	setcolor(UWHITE);
	moveto(2,313);	outtext("Aud Filt");
	moveto(2,323);  outtext("DeNoise");
	moveto(2,333);  outtext("DelayRt");

/*	setcolor(UBEIGE);
	moveto(2,333);  outtext("ANL"); */
/*	moveto(2,243);  outtext(""); */		/* Empty line */

	setcolor(UBEIGE);
	line(1,354,127,354);
/*	moveto(2,358);  outtext("BinAve"); */   /* Not used now */
	setcolor(UWHITE);
	moveto(2,368);	outtext("SpecAve");
	moveto(2,378);	outtext("Window");

	setcolor(UBEIGE);
	line(1,389,127,389);
	setcolor(UWHITE);
	moveto(2,393);	outtext("AF Gain");
	moveto(2,403);	outtext("RF Gain");
	moveto(2,413);	outtext("AGC");
	moveto(2,423);  outtext("AutDisp");

	setcolor(UBEIGE);
	line(1,434,127,434);

	setcolor(UWHITE);
	setviewport(1,1,639,479,DO_CLIP);
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(126, 233);
	sprintf(cc, "%d", contrast);
	outtext(cc);
	moveto(126, 243);
	sprintf(cc, "%d", bright);
	outtext(cc);
	moveto(126, 258);
	sprintf(cc, "%d", mike_gain);
	outtext(cc);
	moveto(126, 268);
	outtext(t[cw_speed].description);
	moveto(126, 278);
	sprintf(cc, "%d", xmit_power);
	outtext(cc);
/*	setcolor(UBEIGE);
	moveto(126, 358);
	sprintf(cc, "%d", 1); */		/* BinAve - not used now */
	outtext(cc);
	setcolor(UWHITE);
	moveto(126, 368);
	sprintf(cc, "%d", noncohave);
	outtext(cc);

	moveto(126, 413);
	sprintf(cc, "%d", rfg[rfgain].dbdisplay);
	outtext(cc);
/*	setcolor(UBEIGE);
	moveto(126, 423);
	sprintf(cc, "%d", 100);
	outtext(cc);
	setcolor(UWHITE);
*/

	if(!waterfall)   draw_panel();
	sa_graph();			/* Put spectrum analyzer graph up */
	set_db_div(1);			/* Add numbers on left */

	normalize_plot(0);		/* Clear normalization */
	}

/* Draw the upper right portion of the screen for the spectrum analyzer
 * display.  Separated so that this same area can be used for a modal
 * dialog box.  Box outline  and markings only.  Grid is drawn during
 * the plotting of the data.
 */
void sa_graph(void)
	{
	int i, yy, kdb;
	char cc[50];

	setviewport(X1L-12, Y1T, X1R, Y1B, DO_CLIP);	
	setcolor(UBEIGE);
	line(12,         0,       12+X1R-X1L, 0);
	line(12,         Y1B-Y1T, 12+X1R-X1L, Y1B-Y1T);
	line(12,         0,       12,         Y1B-Y1T);
	line(12+X1R-X1L, 0,       12+X1R-X1L, Y1B-Y1T);

	settextstyle(0, 0, 1);
	settextjustify(1,1);
	/* db_div_set:  0=10db/div, 1=5db, 2=2db, 3=1db */
	if (db_div_set==0) kdb=10;
	else if (db_div_set==1) kdb=5;
	else if (db_div_set==2) kdb=2;
	else kdb=1;
	for (i=0; i<=8; i++)
		{
		yy = i*16;
		moveto(0, yy);
		sprintf(cc, "%d", kdb*(8 -i));
		outtext(cc);		/* dB level markings */
		}
	}

void mark_frequency(int fmax)
	{
	char cc[50];
	int i, k;

	setcolor(UBEIGE);
	settextstyle(0, 0, 1);
	settextjustify(1, 2);
	setviewport(178, 134, 638, 144, DO_CLIP);
	clearviewport();
	moveto(430, 0);
	if (fmax<10000)
		{
		k=fmax/4;
		outtext("Hz");
		}
	else
		{
		k=fmax/4000;
		outtext("kHz");
		}
	for(i=0; i<4; i++)
		{
		moveto(57+i*107, 0);
		sprintf(cc, "%d", k*(i+1));
		outtext(cc);			/* Freq markings */
		}
	}

/* draw_panel() is used for the conventional panel display.  The lower
 * waterfall window is divided into two parts, the lower 2/3 for help
 * info and this frequency display in the top 1/3.  This display is never
 * used with the waterfall display.
 */
void draw_panel(void)
	{
	int i, x;

	/* 128, 639   d=511  */
	setviewport(X2L+1, Y2T, X2R-1, Y2T+95, DO_CLIP);
	clearviewport();
	setcolor(UBEIGE);
	setlinestyle(SOLID_LINE, 0, THICK_WIDTH);
	rectangle(50, 2, 461, 55);	/* Big Frequency Box */
	setcolor(UWHITE);
	line(55, 41, 455, 41);
	setlinestyle(SOLID_LINE, 0, NORM_WIDTH);
	line(0, 0, 46,0);  line(0, 3, 46, 3);	/* Trim lines at top */
	line(465, 0, 510, 0);  line(465, 3, 510, 3);

	for(i=0; i<5; i++)
	    {
	    x = 95+i*80;
	    line(x, 43, x, 32);
	    tiny_char(1,   x-9, 51);   /* Print with bottom left at coords */
	    tiny_char(4,   x-3, 51);
	    tiny_char(i+4, x+3, 51);
	    }
	for(i=0; i<25; i++)
	    line(63+i*16, 43, 63+i*16, 37);

	/* Now put the general help screen at the bottom: */
	setcolor(UWHITE);
	settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);

	settextjustify(0,2);		/* Left and top */
	/* Works out to be 224 pixels high: */
	setviewport(X2L+1, Y2T+96, X2R-1, Y2B-12, DO_CLIP);
	getimage(0,  0, 510, 160, big_area_ptr1);
	getimage(0, 161, 510, 224, big_area_ptr2);
	clearviewport();
	line(1, 5, 509, 5);		/* Half into "first" line (blank) */
	line(1, 223, 509, 223);
 /* 63 characters:
          123456789012345678901234567890|23456789012345678901234567890123 */
 moveto(3, 3);			/* Line 1-1 */
 /* Blank */
 moveto(3, 13);			/* Line 1-2 */
 outtext("               DSP-10 KEYBOARD COMMANDS - Summary");
 moveto(3, 23);			/* Line 1-3 */
 outtext("       FREQUENCY              |                 GAIN");
 moveto(3, 33);			/* Line 1-4 */
 outtext("1 Hz   Ctrl-Alt-Sh F9 & F10   |   AF Gain               F7 & F8");
 moveto(3, 43);			/* Line 1-5 */
 outtext("10 Hz Down               F9   |   RF Gain            Sh F7 & F8");
 moveto(3, 53);			/* Line 1-6 */
 outtext("10 Hz Up                 F10  |   Mike Gain      Alt-Sh F7 & F8");
 moveto(3, 63);			/* Line 1-7 */
 outtext("100 Hz           Sh F9 & F10  |");
 moveto(3, 73);			/* Line 1-8 */
 outtext("1 kHz           Alt F9 & F10  |   CW Send Speed     Alt F7 & F8");
 moveto(3, 83);			/* Line 1-9 */
 outtext("10 kHz         Ctrl F9 & F10  |   Require Scroll Lock On:");
 moveto(3, 93);			/* Line 1-10 */
 outtext("100 kHz    Ctrl-Alt F9 & F10  |   AGC Release        Scrl G & H");
 moveto(3, 103);		/* Line 1-11 */
 outtext("1 MHz       Ctrl-Sh F9 & F10  |   Transmit Power     Scrl O & P");
 moveto(3, 113);		/* Line 1-12 */
 outtext("                              |   CW Sidetone Lev    Scrl S & D");
 moveto(3,123);			/* Line 1-13 */
 outtext("FM Squelch        Alt S & D   |   ");
 moveto(3,133);			/* Line 1-14 */
 outtext("Save Screen           Alt G   |");
 moveto(3,143);			/* Line 1-15 */
 outtext("                              |       TRANSMIT/RECEIVE");
 moveto(3, 153);		/* Line 1-16 */
 outtext("RIT On/Off            Alt O   |   T/R Toggle               Home");
 moveto(3, 163);		/* Line 1-17 */
 outtext("RIT Down/Up 10Hz  Alt R & T   |");
 moveto(3, 173);		/* Line 1-18 */
 outtext("RIT Zero              Alt Z   |          AF FILTERS");
 moveto(3, 183);		/* Line 1-19 */
 outtext("Mode change           Alt M   |   Filters On/Off             F3");
 moveto(3, 193);		/* Line 1-20 */
 outtext("Help Screens             F1   |   Filter 1/Filter 2          F4");
 moveto(3, 203);		/* Line 1-21 */
 outtext("Display Format        Alt Y   |   DeNoise On/Off          Sh F3");
 moveto(3, 213);		/* Line 1-22 */
 outtext("Exit to DOS     Scrl-Alt F4   |   De-Noise Gain    Ctrl F3 & F4");
	}


/* morse(cmd, ichar) - CW transmission routines to keep track of
 * the send buffer and display it.  Alphanumeric keyboard entries
 * are sent here during both transmission and reception. Commands:
 * INIT_CW	0	Initialize Morse code routine
 * ADD_CH	1	Put char into cw buffer
 * SEND_CW	2	Send a character if available
 *
 * Two special characters are:
 *	# indicates the beginning and ending of a repeat string
 *	@ indicates to stop sending a repeat string
 * These two characters do not actually send any Morse code, themselves.
 * Only a @ will be accepted during a repeat.
 * The repeat string cancels all current sending activity.
 */
void morse(int cmd, int ichar)
	{
	#define BUFSIZE 179
	static int cwbuff[BUFSIZE],	/* Circular buffer for cw xmit */
		    *addptr,		/* Next location to put ch */
		    *sendptr;		/* Next location to get ch from */
	static int bufcount;		/* Ch left to be sent */
	static int cw_key_stat;		/* CW status based on right-alt key */
	int i, right_alt;
	static int repeat;		/* 0 --> No repeat activity
					   1 --> Entering repeat
					   2 --> Repeat entered and active */
	static int *repeat_endptr;	/* Buff always starts at cwbuff */

	switch (cmd){
	    case INIT_CW:{
		for (i=0; i<BUFSIZE; i++)
		    cwbuff[i] = 0;
		addptr = cwbuff;
		sendptr = cwbuff;
		bufcount = 0;
		cw_key_stat = 0;
		break;
		}
	    case ADD_CH:
		{
		if (repeat==2)
		    {
		    if (ichar==(int)'@')	/* Only key responded to */
			{
			repeat=0;
			sendptr = cwbuff;	/* Fresh start */
			addptr = cwbuff;
			bufcount = 0;
			break;
			}
		    else
			break;		/* Break case */
		    }
		if (repeat==0 && ichar==(int)'#')
		    {
		    repeat = 1;		/* Start of repeat enter */
		    sendptr = cwbuff;
		    addptr = cwbuff;
		    print_medium('#', 255); /* Print # on screen */
		    break;
		    }
		if (repeat==1 && ichar==(int)'#')
		    {
		    repeat = 2;		/* End of repeat enter */
		    repeat_endptr = addptr-1;
		    print_medium('#', 255); /* Print on screen */
		    break;
		    }
		if (ichar==8)	/* Backspace */
		    {
		    if (bufcount <= 0)  break;	/* Don't backspace empty buf */
		    bufcount--;
		    if (--addptr < cwbuff)
			addptr = cwbuff+(BUFSIZE-1);	/* Circular buffer */
		    print_medium(8, 254);	/* Backspace the display */
		    break;
		    }
		if (bufcount < BUFSIZE)	/* If there is room */
		    {
		    bufcount++;
		    *addptr = ichar;		/* Put char into buffer */
		    /* Last array position is cwbuff+BUFSIZE-1:  */
		    if (++addptr >= cwbuff+BUFSIZE)
			addptr = cwbuff;	/* Circular buffer */
		    print_medium((char)ichar, 255); /* Print on screen */
		    }
		else
		    beep(50);			/* No room */
		break;
		}
	    case SEND_CW:
		{
		right_alt=_bios_keybrd(_NKEYBRD_SHIFTSTATUS) & 0x0800;
		if ( (right_alt > 0) && (cw_key_stat==0) )
		    {
		    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		    ASYNCOUT(17); ASYNCOUT(2);
		    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
		    cw_key_stat=1;
		    break;
		    }
		if ( (right_alt == 0) && (cw_key_stat==1) )
		    {
		    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		    ASYNCOUT(1); ASYNCOUT(2);
		    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
		    cw_key_stat=0;
		    break;
		    }
		if (repeat==1)  break;
		if (repeat==2)
		    {
		    morse_ch((unsigned char)*sendptr, cw_speed); /* CW  char */
		    if (++sendptr > repeat_endptr)
			sendptr = cwbuff;	/* Keep repeating */
		    break;
		    }
		if (bufcount > 0)	/* If there are char to be sent */
		    {
		    bufcount--;
		    morse_ch((unsigned char)*sendptr, cw_speed); /* CW  char */
		    if (++sendptr >= cwbuff+BUFSIZE)
			sendptr = cwbuff;	/* Circular buffer */
		    }
		break;
		}		
	    default: break;
	    }
	}

void delayu(int ncount)
	{int i;
	 for (i=0; i<ncount; i++);
	}
	
void point1(int x, int y)	/* Cross */
	{
	line(x, y-1, x, y+1);
	line(x-1, y, x+1, y);
	}

void point2(int x, int y)	/* Block square */
	{
	line(x-1, y-1, x+1, y-1);
	line(x-1, y, x+1, y);
	line(x-1, y+1, x+1, y+1);
	}

void point3(int x, int y)	/* Diagonal cross */
	{
	line(x-1, y-1, x+1, y+1);
	line(x-1, y+1, x+1, y-1);
	}

void point4(int x, int y)	/* Block square */
	{
	line(x-1, y-1, x+1, y-1);
	line(x-1, y-1, x-1, y+1);
	line(x+1, y-1, x+1, y+1);
	line(x-1, y+1, x+1, y+1);
	}

void point5(int x, int y)	/* Block rectangle */
	{
	line(x-1, y-1, x+2, y-1);
	line(x-1, y-1, x-1, y+1);
	line(x+2, y-1, x+2, y+1);
	line(x-1, y+1, x+2, y+1);
	}

/* tiny_time( ) puts the time along the left side of the spectral map.
 * The area 9 high and 32 wide should be cleared.
 */
void tiny_time(int hhmm, int x0, int y0)
	{
	int h1, h2, m1, m2;
	h1=hhmm/1000;
	hhmm-=h1*1000;
	h2=hhmm/100;
	hhmm-=h2*100;
	m1=hhmm/10;
	m2=hhmm-m1*10;
	tiny_char(h1, x0, y0);
	tiny_char(h2, x0+6, y0);
	tiny_char(m1, x0+12, y0);
	tiny_char(m2, x0+18, y0);
	line(x0+24, y0, x0+30, y0);	/* Tick for actual time */
	}

/* Print n with bottom left at (x0, y0) */
void tiny_char(int n, int x0, int y0)
	{
	if (n<0 || n>9) return;
	switch (n)
	  {
	  case 0:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-6, x0+4, y0-6);
		line(x0, y0, x0, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 1:{
		line(x0+2, y0, x0+2, y0-6);
		break;
		}
	  case 2:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0, y0, x0, y0-3);
		line(x0+4, y0-3, x0+4, y0-6);
		break;
		}
	  case 3:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 4:{
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-3, x0, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 5:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-3);
		line(x0, y0-3, x0, y0-6);
		break;
		}
	  case 6:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0, x0, y0-6);
		line(x0+4, y0, x0+4, y0-3);
		break;
		}
	  case 7:{
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 8:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0, y0, x0, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 9:{
		line(x0, y0-3, x0, y0-6);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  }
	}

void bar_graph(int x0, int y0, unsigned int value)
	{
	int i, color_on;
	unsigned int test_value;
	static int bar_color[]={UWHITE, UWHITE, UWHITE, UWHITE,
				UWHITE, UWHITE, UOLIVE, UOLIVE,
				UOLIVE, UOLIVE, UOLIVE, UOLIVE,
				ULIGHTRED, ULIGHTRED, ULIGHTRED};
	setfillstyle(SOLID_FILL, UGREENBLACK);
	test_value = 0X4000;
	for(i=14; i>=0; i--)
		{
		color_on = ( value >= test_value );
		test_value /= 2;
		if (color_on)
			setfillstyle(SOLID_FILL, bar_color[i]);
		bar(x0+i*6, y0, x0+i*6+4, y0+8);
		}
	}

void beep(int tau)
	{sound(600);
	delay(tau);
	nosound();
	}

/* Diagnostics usually uses the bottom line on the display to display info
 * about the inner workings of the transceiver.
 * Apr99 Added Moon/Sun track as 'diagnostic' option.
 * The data depends on the value of diag_mode:
 *	0  No diagnostic info is shown.
 *	1  The 20 bytes of status_buf data are printed in hex.
 *	2  Two bar graphs of A/D lvel and D/A level are shown.
 *	3  Moon data
 *	4  Sun data
 */
void diagnostics(int do_what)
	{
	char cc[50];
	int i;

	if (do_what==0)			/* do_what==0 for toggle diag_mode */
		{
		diag_mode++;

		if ((diag_mode) > 4)  diag_mode=0;

		setviewport(129, 470, 639, 479, DO_CLIP); /* Clear main line */
		clearviewport();
		settextjustify(0,2);
		moveto(10, 0);
		sprintf(cc, "%.1d", diag_mode);
		outtext(cc);

		setviewport(1, 470, 127, 479, DO_CLIP); /* Clear left line */
		clearviewport();
		if (diag_mode==3 || diag_mode==4)
		    set_cur_loc(-1);		/* Update screen at bottom */
		return;
		}
	/* do_what==1 for display according to diag_mode */
	if (diag_mode>0)
	  {
	  setviewport(0, 470, 127, 479, DO_CLIP);
	  clearviewport();
	  settextstyle(0, 0, 1);
	  settextjustify(0, 2);
	  moveto(0, 0);
	  sprintf(cc, "%d", ASYNCINSTAT() );  /* Always put loading info out */
	  outtext(cc);
	  moveto(100, 0);
	  sprintf(cc, "%d", incount);	/*  and the number of bytes of data */
	  outtext(cc);
	  settextjustify(1,2);
	  moveto(64, 0);
	  sprintf(cc, "%d", dsp_status[15]);  /* and the fft scaling factor */
	  outtext(cc);
	  settextjustify(0,2);
	  }
	if (diag_mode==0 || diag_mode>=3)  return;

	if (diag_mode==1)
	    {setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	     clearviewport();
	     settextjustify(0,2);
	     for(i=0; i<18; i++)
		{
		moveto(10+i*20, 0);
		sprintf(cc, "%2.2X", dsp_status[i]);
		outtext(cc);
		}
	    moveto(373, 0);
	    if (transmit)
		outtext("T");
	    else
		outtext("R");
	    moveto(385, 0);
	    outtext("Ver1.50");

	    /* This is disabled because there is no data here to show:
	    if(transmit)
		{
		setviewport(150, 30, 450, 50, DO_CLIP);
		clearviewport();
		settextstyle(0, 0, 1);
		settextjustify(0, 2);
		for(i=0; i<10; i++)
		    {
		    moveto(10+i*20, 0);
		    sprintf(cc, "%2.2X", dsp_xmit_status[i]);
		    outtext(cc);
		    if (4*(i/4)==i)  outtext(" ");
		    }
		}
	    */

	    if ((!transmit) && auto_disp_on)
		{moveto(400, 0);
		 if (ave_spectral_pwr<1e-12)
			ave_spectral_pwr=1e-12;
		 sprintf(cc, "%7.2lf", 10.0*log10(ave_spectral_pwr));
		 outtext(cc);
		}
	    }
	else if (diag_mode==2)
	    {
	    setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	    clearviewport();

	    moveto(10, 0);
	    outtext("A/D Level:");
	    bar_graph( 100, 0, 256*dsp_status[10]+dsp_status[9] );

	    moveto(220, 0);
	    outtext("D/A Level:");
	    bar_graph( 300, 0, 256*dsp_status[12]+dsp_status[11] );

	    if ((!transmit) && auto_disp_on)
		{moveto(400,0);
		 if (ave_spectral_pwr<1e-12)
			ave_spectral_pwr=1e-12;
		 sprintf(cc, "%7.2lf", 10.0*log10(ave_spectral_pwr));
		 outtext(cc);
		}
	    }
	}

/* Preliminaries and tables */
void initialize(void)
	{
	int i, ifg;
	double fg;
	long  a, b, fr;
	unsigned int xu;

	ASYNCINIT(com_port);
	ASYNCSET(BAUD_RATE, COM_SETTINGS);
	ASYNCCLEAR();

	/* Following vector converts (0,255), representing 0 to
	    80 dB into a power ratio (1.0, 10^8).  Use vector for speed */
	for (i=0; i<256; i++)		/* Convert log to power */
		power10[i]=pow(10.0, 0.03125 * (double)i );

	/* Gain function in 1 dB steps for use in dsp */
	for(i=1; i<=100; i++)		/* Setup gain fcn as 2 bytes */
		{fg = pow(10.0, (double)(i-100) / 20. );
		 fgain[i] = fg;
		 ifg = near_int(32767.0*fg);
		 fgainl[i] = ifg & 0XFF;
		 fgainh[i] = (ifg >> 8) & 0XFF;
		}

	for(i=0; i<515; i++)  dnorm[i]=0.0;

	do_warm_boot();     	/* Fully set DSP & screen to current state */

	/* Setup reference divider in synthesizer */
	send_shift_data(R10MHZ);	/* R1MHZ, R5MHZ, R6_4MHZ or R10MHZ */
	delay(250);

	if (!waterfall)
	    {
	    draw_panel();
	    setviewport(0,124,127,161,DO_CLIP);	/* Small Freq Box */
	    clearviewport();
	    }

	if(input_if_af==0) new_band(band);

	setviewport(0,60,101,70,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	setcolor(UWHITE);
	moveto(50,1);
	if (input_if_af==0)
		outtext("RECEIVE");
	else
		outtext("AUDIO");
	print_big(0, -1, -1);		/* Initialize scroll printing */
	print_medium(0, -1);
	morse(INIT_CW, 0);
	set_contrast();
	set_brightness();
	contrast_line_count=0;		/* Don't show "Contrast Line" */
	do_auto_disp(0);		/* Start with auto display off */

	new_freq();			/* Initialize freq */

	/* Following is a kludge to make  sure that hardware is properly
	 * properly initialized.  Someday find out why needed!!
	 */
	tfreqL -= 1000000;
	new_freq();
	delay(10);
	tfreqL += 1000000;
	new_freq();
	}

/* The following preset values are entered for now, but may be overwritten
 * by the .CFG file if the associated entries are present.
 * The tvtr[] structure provides one preset for each of the 16 VHF/UHF/uWave
 * bands plus 5 more 'Memory' channels that can be used on any band by the
 * UHFA.CFG file.  These are invoked by Ctrl-Alt x where x is either a band
 * from A to P or a memory channel 1 to 5.
 */
void presets(void)
	{
	int i;

	strcpy(tvtr[0].name, "50 MHz");
	tvtr[0].lo_type=0;
	tvtr[0].sb=-1;
	tvtr[0].lo_freq=-94000000.0;		/* - sign for upconversion */
	tvtr[0].chain_mult=1;
	strcpy(tvtr[0].init_str, "  ");
	tvtr[0].gain_db=20.0;
	tvtr[0].control=0;
	tvtr[0].bus_adr=0X00;
	tvtr[0].last_tfreq=144125000.0;
	tvtr[0].last_mode=USB;

	/* 2 Meters is a dummy transverter with LO frequency of 0.0 Hz */
	strcpy(tvtr[1].name, "144 MHz");
	tvtr[1].lo_type=0;
	tvtr[1].sb=-1;
	tvtr[1].lo_freq=0.0;
	tvtr[1].chain_mult=1;
	strcpy(tvtr[1].init_str, "  ");
	tvtr[1].gain_db=0.0;
	tvtr[1].control=0;
	tvtr[1].bus_adr=0X10;
	tvtr[1].last_tfreq=144200000.0;
	tvtr[1].last_mode=USB;

	for (i=2; i<9; i++)
	    tvtr[i].lo_type=999;		/* Transverter not installed */

	strcpy(tvtr[9].name, "10 GHz");
	tvtr[9].lo_type=0;
	tvtr[9].sb=-1;
	tvtr[9].lo_freq=10224000000.0;
	tvtr[9].chain_mult=1;
	strcpy(tvtr[9].init_str, "  ");
	tvtr[9].gain_db=28.0;
	tvtr[9].control=0;
	tvtr[9].bus_adr=0X20;
	tvtr[9].last_tfreq=144100000.0;
	tvtr[9].last_mode=USB;

	for (i=10; i<15; i++)
	    tvtr[i].lo_type=999;		/* Transverter not installed */

	strcpy(tvtr[15].name, "15 kHz Input");
	tvtr[15].lo_type=0;
	tvtr[15].sb=-1;
	tvtr[15].lo_freq=-143085000.0;
	tvtr[15].chain_mult=1;
	strcpy(tvtr[15].init_str, "  ");
	tvtr[15].gain_db=17.0;
	tvtr[15].control=0;
	tvtr[15].bus_adr=0X10;
	tvtr[15].last_tfreq=144000000.0;
	tvtr[15].last_mode=USB;

	strcpy(tvtr[16].name, "144MHz VFO 1");
	tvtr[16].lo_type=0;
	tvtr[16].sb=-1;
	tvtr[16].lo_freq=0.0;
	tvtr[16].chain_mult=1;
	strcpy(tvtr[16].init_str, "  ");
	tvtr[16].gain_db=0.0;
	tvtr[16].control=0;
	tvtr[16].bus_adr=0X10;
	tvtr[16].last_tfreq=144100000.0;
	tvtr[16].last_mode=CW;

	strcpy(tvtr[17].name, "144MHz VFO 2");
	tvtr[17].lo_type=0;
	tvtr[17].sb=-1;
	tvtr[17].lo_freq=0.0;
	tvtr[17].chain_mult=1;
	strcpy(tvtr[17].init_str, "  ");
	tvtr[17].gain_db=0.0;
	tvtr[17].control=0;
	tvtr[17].bus_adr=0X10;
	tvtr[17].last_tfreq=144200000.0;
	tvtr[17].last_mode=USB;

	strcpy(tvtr[18].name, "144MHz VFO 3");
	tvtr[18].lo_type=0;
	tvtr[18].sb=-1;
	tvtr[18].lo_freq=0.0;
	tvtr[18].chain_mult=1;
	strcpy(tvtr[18].init_str, "  ");
	tvtr[18].gain_db=0.0;
	tvtr[18].control=0;
	tvtr[18].bus_adr=0X10;
	tvtr[18].last_tfreq=144250000.0;
	tvtr[18].last_mode=USB;

	strcpy(tvtr[19].name, "144MHz VFO 4");
	tvtr[19].lo_type=0;
	tvtr[19].sb=-1;
	tvtr[19].lo_freq=0.0;
	tvtr[19].chain_mult=1;
	strcpy(tvtr[19].init_str, "  ");
	tvtr[19].gain_db=0.0;
	tvtr[19].control=0;
	tvtr[19].bus_adr=0X10;
	tvtr[19].last_tfreq=146580000.0;
	tvtr[19].last_mode=FM;

	strcpy(tvtr[20].name, "144MHz VFO 5");
	tvtr[20].lo_type=0;
	tvtr[20].sb=-1;
	tvtr[20].lo_freq=0.0;
	tvtr[20].chain_mult=1;
	strcpy(tvtr[20].init_str, "  ");
	tvtr[20].gain_db=0.0;
	tvtr[20].control=0;
	tvtr[20].bus_adr=0X10;
	tvtr[20].last_tfreq=146940000.0;
	tvtr[20].last_mode=FM;

	/* Initialize the locations to non-existant except 0 which is at
	 * lat=0 and lon=0.  Non-existant is time_zone=999.
	 */
	for (i=0; i<=9; i++)
	    {
	    loc[i].lat=0.0;
	    loc[i].lon=0.0;
	    loc[i].ht=0.0;
	    loc[i].time_zone=999;
	    loc[i].dst=0;
	    strcpy(loc[i].nm, "");
	    strcpy(loc[i].name, "");
	    }
	loc[0].time_zone=0;
	strcpy(loc[0].nm, "(0,0)");
	strcpy(loc[0].name, "Lat=0 Lon=0");
	}

void print_time(void)
	{int i;
	 time_t t0;
	 time_t tdt;			/* ut+deltat */
	 time_t deltat;
	 struct tm *tp;
	 struct tm *ptdt;		/* tdt=ut+deltat */
	 double y2;			/* cal year - 2000 */
	 char cc[40];

	 time(&t0);
	 if (time_type)			/* time_type==1 --> use local time */
	    tp=localtime(&t0);
	 else
	    tp=gmtime(&t0);		/* time_type==0 --> use GMT */
	 strcpy(cc, asctime(tp));	/* Rearrange the characters */
	 cc[0]=cc[8]; cc[1]=cc[9]; cc[2]=32;	/* Get Date */
	 cc[3]=cc[4]; cc[4]=cc[5]; cc[5]=cc[6]; cc[6]=32;
	 cc[7]=cc[22]; cc[8]=cc[23]; cc[9]=0;
	 setviewport(0,437,127,457,DO_CLIP);
	 clearviewport();
	 setcolor(UWHITE);
	 settextstyle(0, 0, 1);
	 settextjustify(1, 2);
	 moveto(63,1);
	 outtext(cc);
	 /* Now create a string with 'hhmm:ss XXX' where XXX is GMT, etc. */
	 cc[0]=cc[11];  cc[1]=cc[12];	/* Get hh */
	 for(i=14; i<=18; i++)		/* Get mm:ss */
	     cc[i-12]=cc[i];
	 cc[7]=32;			/* Space */
	 for (i=0; i<=3; i++)
	    cc[i+8]=time_str[i];	/* XXX */
	 cc[12]=0;			/* End of string */
	 moveto(63,11);
	 outtext(cc);

	/* Now update Moon &/| Sun data, Process only if GMT is available */
	if(ut_available)		/* We can update Moon/Sun data */
	    {
	    /* We find deltat by knowing the year.  From this we get tdt:  */
	    y2=(double)(tp->tm_year-100)+(double)tp->tm_mon/12.0;
	    /* Fit to data as of 1999:  */
	    deltat=(time_t) near_int(63.86+1.037*y2);
	    tdt=t0+deltat;
	    ptdt=gmtime(&tdt);

	    if (diag_mode==3)
		update_moon(ptdt, 0);
	    else if (diag_mode==4)
		update_sun(ptdt, 0);
	    }
	}

void update_moon(struct tm *ptm, int do_now)
    {
    int j;
    char cc[100];
    static int last_min;
    static int last_day;
    struct latlon ll;
    struct radec  rd;
    struct hadec  hd, hd1;
    struct azel   ae;
    double gst;
    double pm;
    double drdt;
    double dop0, dop1;

    if ((ptm->tm_yday != last_day) || do_now)
	{
	last_day=ptm->tm_yday;	/* Won't do this for another day */
	cur_env.day=(double) ptm->tm_mday;
	cur_env.month=1+ptm->tm_mon;
	cur_env.year=1900+ptm->tm_year;
	julday(&cur_env);		/* Update Julian day */
	nutat(&cur_env);		/* Once a day find nutation */
	obliq(&cur_env);		/*  and obliquity of ecliptic */
	}
    if ((last_min != ptm->tm_min) || do_now)
	{
	last_min=ptm->tm_min;	/* Don't do this for another min */
	gst=cnvlt2gst( (double)ptm->tm_hour +
	    ((((double)ptm->tm_min) + ((double)ptm->tm_sec/60.0))/60.0),
	    &cur_env, &loc[0], &ut);
	moon(cur_env.dj, ut, &ll.lon, &ll.lat, &pm, &drdt);
	cnvll2radec(&ll, &rd, &cur_env);

	/* We calculate doppler twice, once for home loc[0] and
	 * then for loc[cur_loc] to get dx location doppler */
	j=cur_loc;
	hd.ha=cnvra2ha(rd.ra, gst, loc[j].lon);
	hd.dec=rd.dec;
	parallax(loc[j].lat, &hd, pm, &hd1);
	cnvhadec2azel(&hd1, &loc[j], &ae);
	/*    dop0=doppler(drdt, &ae, (double)tfreqL, loc[j].lat);  */
	if (input_if_af)		/* Audio processor, use f_eme */
	    dop1=doppler(drdt, &ae, f_eme, loc[j].lat);
	else				/* Full radio, use tfreqL */
	    dop1=doppler(drdt, &ae, (double)tfreqL, loc[j].lat);

	j=0;				/* Now for home location */
	hd.ha=cnvra2ha(rd.ra, gst, loc[j].lon);
	hd.dec=rd.dec;
	parallax(loc[j].lat, &hd, pm, &hd1);
	cnvhadec2azel(&hd1, &loc[j], &ae);
	/*    dop0=doppler(drdt, &ae, (double)tfreqL, loc[j].lat);  */
	if (input_if_af)		/* Audio processor, use f_eme */
	    dop0=doppler(drdt, &ae, f_eme, loc[j].lat);
	else				/* Full radio, use tfreqL */
	    dop0=doppler(drdt, &ae, (double)tfreqL, loc[j].lat);

	setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	clearviewport();
	settextjustify(0,2);
	/* In following, for rel dB, the nominal pm, in radians is
	 * d2r(57.3/60).  1.0 over this is 59.9934 */
	sprintf(cc, "Moon: %s-%s AZ=%.2lf EL=%.2lf %dHz %.2lfdB",
	    loc[0].nm, loc[cur_loc].nm, r2d(ae.az), r2d(ae.el),
	    near_int(dop0+dop1), 40.0*log10(59.9934*pm));
	moveto(0, 0);
	outtext(cc);
	}
    }

/* update_sun()  -  If the diag_mode==4 this will print the az and el of the
 * Sun on the bottom diagnostic line.  Earth location is loc[0] always.
 * Also slight variation in Sun intensity due to distance != 1 AU.
 */
void update_sun(struct tm *ptm, int do_now)
    {
    char cc[100];
    static int last_min;
    static int last_day;
    struct latlon ll;
    struct radec  rd;
    struct hadec  hd, hd1;
    struct azel   ae;
    double gst;
    double sun_dist;

    if ((ptm->tm_yday != last_day) || do_now)
	{
	last_day=ptm->tm_yday;	/* Won't do this for another day */
	cur_env.day=(double) ptm->tm_mday;
	cur_env.month=1+ptm->tm_mon;
	cur_env.year=1900+ptm->tm_year;
	julday(&cur_env);		/* Update Julian day */
	nutat(&cur_env);		/* Once a day find nutation */
	obliq(&cur_env);		/*  and obliquity of ecliptic */
	}
    if ((last_min != ptm->tm_min) || do_now)
	{
	last_min=ptm->tm_min;	/* Don't do this for another min */
	gst=cnvlt2gst( (double)ptm->tm_hour +
	    ((((double)ptm->tm_min) + ((double)ptm->tm_sec/60.0))/60.0),
	    &cur_env, &loc[cur_loc], &ut);
	sun(cur_env.dj, ut, &ll.lon, &sun_dist);
	ll.lat=0.0;			/* Easy for the Sun */
	/* Now a couple of minor corrections.  These are not generally
	 * needed.  They maintain an accuracy of better than 1 sec in
	 * angle.  First for nutation from cur_env.dp [from nutat() ].
	 * Then a fixed correction of 0.00569 deg that is the change in Sun
	 * position during the time that the light was travelling to the Earth,
	 * called aberation.
	 */
	ll.lon += (cur_env.dp-d2r(0.00569));
	cnvll2radec(&ll, &rd, &cur_env);	/* Get RA and Dec */
	/* Convert RA to hour angle: */
	hd.ha=cnvra2ha(rd.ra, gst, loc[cur_loc].lon);
	hd.dec=rd.dec;
	/* A tiny correction for parallax (we are not at Earth center):  */
	parallax(loc[cur_loc].lat, &hd, d2r(0.0024428)/sun_dist, &hd1);
	cnvhadec2azel(&hd1, &loc[cur_loc], &ae);
	    
	setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	clearviewport();
	settextjustify(0,2);
	sprintf(cc, "Sun: %s AZ=%.2lf EL=%.2lf %.2lfdB",
	    loc[cur_loc].nm, r2d(ae.az), r2d(ae.el), -20.0*log10(sun_dist));
	moveto(0, 0);
	outtext(cc);
	}
    }

/* set_cur_loc()  -  Not all loc[i] need exist.  This moves through  to
 * those that do.  Note the data is initialized so that loc[0] always
 * exists so this routine can't hang. loc[i].time_zone is not used in
 * the program now except to keep track of the active loc[i].
 * If updn==-1 there is no change in loc[i], but just update screen.
 */
void set_cur_loc(int updn)
    {
    time_t t0;
    time_t tdt, deltat;
    struct tm *tp, *ptdt;
    double y2;

    if (updn==1)				/* Go to next */
	{
	if (++cur_loc > 9)  cur_loc=0;
	while (loc[cur_loc].time_zone == 999)
	    {
	    if (++cur_loc > 9)  cur_loc=0;
	    }
	}
    else if (updn==0)				/* Go to previous */
	{
	if (--cur_loc < 0)  cur_loc=9;
	while (loc[cur_loc].time_zone == 999)
	    {
	    if (--cur_loc < 0)  cur_loc=9;
	    }
	}
    /* Now update bottom line of screen */
    if (ut_available)
	{
	time(&t0);
	tp=gmtime(&t0);
	y2=(double)(tp->tm_year-100)+(double)tp->tm_mon/12.0;
	/* Fit to data as of 1999:   */
	deltat=(time_t) near_int(63.86+1.037*y2);
	tdt=t0+deltat;
	ptdt=gmtime(&tdt);
	if (diag_mode==3)
	    update_moon(ptdt, 1);	/* 1 for 'do_now'  */
	else if (diag_mode==4)
	    update_sun(ptdt, 1);
	}
    if(!ut_available && (diag_mode==3 || diag_mode==4))
	{
	setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	clearviewport();
	settextjustify(0,2);
	moveto(0, 0);
	outtext("Moon/Sun data unavailable; need TZ in DOS enviornment");
	}
    }

/* Hardware specific code to load both the LMX1501 synthesizer as well
 * as the 16 bit control shift register.  lsynth is a 19 bit data
 * word to program the synthesizer.  It comes in two forms, one to
 * set up the ref divider and another to set the frequency.  Both are treated
 * identically here.
 *
 * The 19+16=35 bits are sent out as three 16 bit commands with the first 13
 * bits being passed through the shift registers:
 *	word_first	MS 3 bits of synthesizer data
 *	word_mid	LS 16 bits of synthesizer data
 *	word_last	All of data destined for the two HC595 shift registers
 * Only 16 bits are sent per command so as to ensure that the shift reg is
 * loaded before the next command is sent asking for more.
 */
void send_shift_data(unsigned long lsynthx)
	{int i;
	 unsigned int  sr;
	 unsigned int  word_first, word_mid;

	 word_first = (unsigned int) (0X7UL & (lsynthx>>16));  /* 3 bits */
	 word_mid   = (unsigned int) (0XFFFFUL & lsynthx);     /* 16 bits */
	 /* 'Word_last' comes from shift_reg (low 8) and
	    rfg[i].gainbyte (high 8)                  */

	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(8);  /* Clock in first 16 bits */
	 ASYNCOUT(word_first & 0XFF); ASYNCOUT((word_first >> 8) & 0X00FFU);
	 ASYNCOUT(3); ASYNCOUT(4);	/* Filler bytes */
 /*  printf("%lu %lu ", (word_first & 0XFF), ((word_first >> 8) & 0X00FFU) ); */
 /* printf("%lX\n", lsynthx); */

	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(8);  /* Clock in middle 16 bits */
	 ASYNCOUT( word_mid & 0XFF ); ASYNCOUT( (word_mid >> 8) & 0X00FFU);
	 ASYNCOUT(3); ASYNCOUT(4);

 /* printf("%lu %lu\n", (word_mid & 0XFF), ((word_mid >> 8) & 0X00FFU) ); */


	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(8);  /* Clock in last 16 bits */
	 ASYNCOUT( shift_reg & 0XFF );
	 ASYNCOUT( rfg[rfgain].gainbyte & 0XFF );
	 ASYNCOUT(3); ASYNCOUT(4);

	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(7);  /* The 7 cmd sets the enable gate*/
	 ASYNCOUT(1); ASYNCOUT(2);	    /*   to load the Shift Reg */
	 ASYNCOUT(dsp_deltap_lo); ASYNCOUT(dsp_deltap_hi); /* and dsp freq */

	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(7);  /* Close the enable gate to */
	 ASYNCOUT(0); ASYNCOUT(2);	    /*   prepare for next time */
	 ASYNCOUT(dsp_deltap_lo); ASYNCOUT(dsp_deltap_hi);
	}

/* do_error(err, tau)  -  Displays error err for about 15 seconds or
 * until replaced by next error.  Sounds beep for tau milliseconds.
 * Returns current error number.  If err==0 no error is set, but
 * current error number or 0 is returned.  err==err+1000 clears
 * err if the current error is err.   err==9999 clears all errors.
 */
int do_error(int err, int tau)
	{
	static int current_err=0;
	char cc[100];    /* cc[15];*/
#if 0
TO USE THIS check_sum_pc and _dsp must be made global<<<<<<<<<<<<<<
 int i;
 static int jj=-1;
 if ((jj++) > 15) jj=-1;
 setviewport(129, 250+10*jj, 639, 259+10*jj, DO_CLIP); /* Clear line */
 clearviewport();
 settextjustify(0,2);
 for(i=0; i<18; i++)
   {
   moveto(10+i*20, 0);
   sprintf(cc, "%2.2X", dsp_status[i]);
   outtext(cc);
   if (4*(i/4)==i)  outtext(" ");
   }
 moveto(410, 0);
 sprintf(cc, "%2.2X", check_sum_pc);
 outtext(cc);
 moveto(430, 0);
 sprintf(cc, "%2.2X", check_sum_dsp);
 outtext(cc);
 moveto(470, 0);
 sprintf(cc, "%d", err);
 outtext(cc);
#endif

	if (beep_on_error)  beep(tau);
	if ((err>0) && (err<1000))
	    display_error_count=25;	/* About 15 seconds */

	if (err==0) return current_err;
	if ((err==9999) || ((err>1000) && (current_err==(err-1000))))
	    current_err=0;
	if (err<1000)
	    current_err=err;
	setviewport(0,457,127,467,DO_CLIP);
	clearviewport();
	setcolor(ULIGHTRED);
	settextstyle(0, 0, 1);
	settextjustify(1, 2);
	moveto(63,1);
	sprintf(cc, "Error %d", current_err);
	if (current_err>0)  outtext(cc);
	setcolor(UWHITE);
	return current_err;
	}

/* near_int(x)  -  Returns the integer nearest x.
 */
int near_int(double x)
	{
	if (x <= -32768.0)  return -32768;
	if (x >= 32767.0)  return 32767;
	if (x >= 0.0)  return (int)(x+0.5);
	return (int)(x-0.5);
	}

int gpixel(int x, int y)
	{
	return getpixel(x, y);
	}

/* get_kbd_int()  -  Returns an integer from keyboard entry.  Assumes text
 * mode (not graphics.)  Ends on 'Enter' and responds to Backspace
 */
int  get_kbd_int(void)
	{
	char cc[50], c, *cptr;
	cptr=cc;
	for(;;)
	    {
	    if (kbhit()) 
		{
		c=(char)(_bios_keybrd(_KEYBRD_READ) & 0XFF);
		if ( (c==8) && (cptr>cc) )
		    {
		    putch(8);
		    putch(32);
		    putch(8);
		    cptr--;
		    }
		else if (c==13)
		    {
		    *cptr=0;
		    putch(13);
		    putch(10);
		    break;
		    }
		else if (c>=32 && c<127)
		    {
		    putch((int)c);
		    *cptr++=c;
		    }
		if ( (cptr-cc)>49 )
		    {
		    *cptr=0;
		    break;
		    }
		}
	    }
	return atoi(cc);
	}	

/* keyident() - Utility that is available before graphics is started.
 * Gives information on key that has been hit, for use in setting up
 * header file for keyboard definition.
 */
int keyident(void)
	{
	unsigned int stat, s, ichr;
	long lkey;

	stat = _bios_keybrd(_KEYBRD_SHIFTSTATUS);
	s = stat;
	/* Make left and right shift keys equals (bits 0 and 1) and
	 * mask out Caps Lock, Num Lock and Insert (bits 5, 6, and 7).     */
	if (stat&2)   stat |= 1;
	stat &= 0X1D;
	if ((stat&1) > 0) printf("SHIFT, ");
	if ((stat&4) > 0) printf("CTRL, ");
	if ((stat&8) > 0) printf("ALT, ");
	if ((stat&16) > 0) printf("ScrlLk, ");
	ichr = _bios_keybrd(_KEYBRD_READ);
	lkey = (long)ichr + (((long)stat)<<16);
	printf(" ichr=%X  stat=%X.  Hex value for header file: 0X%lX\n",
		ichr, s, lkey);
	if (stat==0 && (ichr&0XFF)==0X0D)
	    return 1;
	else
	    {
	    printf("\nHit 'Enter' to proceed (others give key ident)\n");
	    return 0;
	    }
	}

void get_mem(void)			/* Get memory for screen scroll */
	{
	int hc;

	/* This is part of sign-on and set-up! */
	if ( (big_area_ptr1 = (char *)malloc(50000) ) == NULL)
		{printf("Unable to get memory for bit image (top). ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (big_area_ptr2 = (char *)malloc(50000) ) == NULL)
		{printf("Unable to get memory for bit image (bottom). ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if (startup_detail)
	    {
	    printf("big_area_ptr1 = %p\n", big_area_ptr1);
	    printf("big_area_ptr2 = %p\n", big_area_ptr2);
	    }
	hc=heapcheck();
	if (hc==_HEAPCORRUPT)
	    {
	    printf("Heap has been corrupted\n");
	    exit(0);
	    }
	if (hc==_HEAPEMPTY)
	    {
	    printf("No heap\n");
	    exit(0);
	    }
	if (startup_detail && hc==_HEAPOK)      printf("Heap is verified\n");
	}

void show_help(void)
	{
	char cc[100];
	char cc1[7], cc2[33];
	int i;

	setcolor(UWHITE);
	settextjustify(0,2);		/* Left and top */
	/* Works out to be 224 pixels high: */
	setviewport(X2L+1, Y2T+96, X2R-1, Y2B-12, DO_CLIP);
	getimage(0,  0, 510, 160, big_area_ptr1);
	getimage(0, 161, 510, 224, big_area_ptr2);
	clearviewport();
	line(1, 0, 509, 0);
	line(1, 223, 509, 223);
 /* 63 characters:
          123456789012345678901234567890|23456789012345678901234567890123 */
 moveto(3, 3);			/* Line 1-1 */
 outtext("             DSP-10 KEYBOARD COMMANDS - Screen 1");
 moveto(3, 13);			/* Line 1-2 */
 outtext("       FREQUENCY                                GAIN");
 moveto(3, 23);			/* Line 1-3 */
 outtext("1 Hz Down    Ctrl-Alt-Sh F9   |   AF Gain Down 2dB           F7");
 moveto(3, 33);			/* Line 1-4 */
 outtext("1 Hz Up      Ctrl-Alt-Sh F10  |   AF Gain Up 2 dB            F8");
 moveto(3, 43);			/* Line 1-5 */
 outtext("10 Hz Down               F9   |   RF Gain Down 6 dB       Sh F7");
 moveto(3, 53);			/* Line 1-6 */
 outtext("10 Hz Up                 F10  |   RF Gain Up 6 dB         Sh F8");
 moveto(3, 63);			/* Line 1-7 */
 outtext("100 Hz Down        Shift F9   |   COLD ON/Mk Dwn 1 dB Alt-Sh F7");
 moveto(3, 73);			/* Line 1-8 */
 outtext("100 Hz Up          Shift F10  |   COLD OFF/Mk Up 1 dB Alt-Sh F8");
 moveto(3, 83);			/* Line 1-9 */
 outtext("1 kHz Down           Alt F9   |   Require Scroll Lock On:");
 moveto(3, 93);			/* Line 1-10 */
 outtext("1 kHz Up             Alt F10  |   AGC Release Faster     Scrl G");
 moveto(3, 103);		/* Line 1-11 */
 outtext("10 kHz Down         Ctrl F9   |   AGC Release Slower     Scrl H");
 moveto(3, 113);		/* Line 1-12 */
 outtext("10 kHz Up           Ctrl F10  |   CW Sidetone Down 1 dB  Scrl S");
 moveto(3,123);			/* Line 1-13 */
 outtext("100 kHz Down    Ctrl-Alt F9   |   CW Sidetone Up 1 dB    Scrl D");
 moveto(3,133);			/* Line 1-14 */
 outtext("100 kHz Up      Ctrl-Alt F10  |");
 moveto(3,143);			/* Line 1-15 */
 outtext("1 MHz Down       Ctrl-Sh F9   |       TRANSMIT/RECEIVE");
 moveto(3, 153);		/* Line 1-16 */
 outtext("RIT 10Hz Down     Alt R or r  |   T/R Toggle               Home");
 moveto(3, 163);		/* Line 1-17 */
 outtext("RIT 10Hz Up       Alt T or t  |");
 moveto(3, 173);		/* Line 1-18 */
 outtext("RIT On/Off        Alt O or o  |          AF FILTERS");
 moveto(3, 183);		/* Line 1-19 */
 outtext("RIT Zero          Alt Z or z  |   Filters On/Off             F3");
 moveto(3, 193);		/* Line 1-20 */
 outtext("Require Scroll Lock on:       |   Filter 1/Filter 2          F4");
 moveto(3, 203);		/* Line 1-21 */
 outtext("CW Offset Down        Scrl C  |   DeNoise On/Off          Sh F3");
 moveto(3, 213);		/* Line 1-22 */
 outtext("CW Offset Up          Scrl V  |   De-Noise Gain    Ctrl F3 & F4");

	for(;;)			/* Wait for keyboard hit */
	    if ( kbhit() )
		{ _bios_keybrd(_KEYBRD_READ);  break; }
	clearviewport();
	line(1, 0, 509, 0);
	line(1, 223, 509, 223);
 /* 63 characters:
          123456789012345678901234567890|23456789012345678901234567890123 */
 moveto(3, 3);			/* Line 2-1 */
 outtext("             DSP-10 KEYBOARD COMMANDS - Screen 2");
 moveto(3, 13);			/* Line 2-2 */
 outtext("          FM                                      OTHER");
 moveto(3, 23);			/* Line 2-3 */
 outtext("SUN ON/Sqlch Dwn Alt S or s   |   Spectral Dsply Ave Dwn Alt F3");
 moveto(3, 33);			/* Line 2-4 */
 outtext("SUN OFF/Sqch Up  Alt D or d   |   Spectral Dsply Ave Up  Alt F4");
 moveto(3, 43);			/* Line 2-5 */
 outtext("                              |   Norm Spectrum Fast Alt Q or q");
 moveto(3, 53);			/* Line 2-6 */
 outtext("          CW                  |   Norm Spectr 10 min Alt P or p");
 moveto(3, 63);			/* Line 2-7 */
 outtext("Alphanumeric  Any key in CW   |   Spectr Norm On/Off Alt N or n");
 moveto(3, 73);			/* Line 2-8 */
 outtext("Delete Last Ch    Backspace   |   Color Display AGC  Alt C or c");
 moveto(3, 83);			/* Line 2-9 */
 outtext("Delete Buffer        Sh DEL   |   Spectr Disp Width       Alt %");
 moveto(3, 93);			/* Line 2-10 */
 outtext("CW Speed Slower      Alt F7   |   Spectr Disp dB/div      Alt !");
 moveto(3, 103);		/* Line 2-11 */
 outtext("CW Speed Faster      Alt F8   |   Spectr Disp Down 1 dB   Alt @");
 moveto(3, 113);		/* Line 2-12 */
 outtext("AR                      '['   |   Spectr Display Up 1 dB  Alt #");
 moveto(3,123);			/* Line 2-13 */
 outtext("SK                      ']'   |   Input RF/Audio       Alt I, i");
 moveto(3,133);			/* Line 2-14 */
 outtext("AS                      '\\'   |   Screen to Disk       Alt G, g");
 moveto(3,143);			/* Line 2-15 */
 outtext("CW Repeat Delimiter     '#'   |   Spectrum to Disk     Alt F, f");
 moveto(3, 153);		/* Line 2-16 */
 outtext("CW Repeat Break         '@'   |   Mode=cw, usb, lsb... Alt M, m");
 moveto(3, 163);		/* Line 2-17 */
 outtext("                              |");
 moveto(3, 173);		/* Line 2-18 */
 outtext("        BAND CHANGE           |   Require scroll lock:");
 moveto(3, 183);		/* Line 2-19 */
 outtext("2-meters   Ctrl-Alt 1,2...5   |   Xmit Power down 1 dB   Scrl O");
 moveto(3, 193);		/* Line 2-20 */
 outtext("50 MHz Trnsvrtr  Ctrl-Alt A   |   Xmit Power Up 1 dB     Scrl P");
 moveto(3, 203);		/* Line 2-21 */
 outtext("144:B  222:C  432:D   903:9   |   Change Diagnostics    Scrl F3");
 moveto(3, 213);		/* Line 2-22 */
 outtext("1296:E  2304:F ... 241GHz:O   |   Exit to DOS       Scrl-Alt F4");

	for(;;)			/* Wait for keyboard hit */
	    if ( kbhit() )
		{ _bios_keybrd(_KEYBRD_READ);  break; }
	clearviewport();
	line(1, 0, 509, 0);
	line(1, 223, 509, 223);
 /* 63 characters:
          123456789012345678901234567890|23456789012345678901234567890123 */
 moveto(3, 3);			/* Line 3-1 */
 outtext("             DSP-10 KEYBOARD COMMANDS - Screen 3");
 moveto(3, 13);			/* Line 3-2 */
 outtext("    SPECTRAL WATERFALL                        OTHER");
 moveto(3, 23);			/* Line 3-3 */
 outtext("Brightness Down          F5   |   Require Scroll Lock on:");
 moveto(3, 33);			/* Line 3-4 */
 outtext("Brightness Up            F6   |   Disp Dot Type         Scrl F5");
 moveto(3, 43);			/* Line 3-5 */
 outtext("Contrast Down         Sh F5   |   Spectral Data 8/16    Scrl F2");
 moveto(3, 53);			/* Line 3-6 */
 outtext("Contrast Up           Sh F6   |   Wrm Boot DSP Scrl-Ctrl-Alt F4");
 moveto(3, 63);			/* Line 3-7 */
 outtext("");
 moveto(3, 73);			/* Line 3-8 */
 sprintf(cc,
         "Change Location     < and >   |   Fdoppler=%0.1lf Hz", f_eme);
 outtext(cc);
 moveto(3, 83);			/* Line 3-9 */
 outtext("             -----  MOON/SUN LOCATIONS  -----");
 for(i=0; i<=9; i++)
    {
    if (loc[i].time_zone != 999)
	{
	sprintf(cc, "%d %6.6s %31.31s %10.5lf %10.5lf",
	    i,
	    left_justify(loc[i].nm, cc1, 6, SPACE),
	    left_justify(loc[i].name, cc2, 31, HYPHEN),
	    r2d(loc[i].lat), r2d(loc[i].lon));
	moveto(3, 93+10*i);
	outtext(cc);
	}
    else
	{
	sprintf(cc, "%d", i);
	moveto(3, 93+10*i);
	outtext(cc);
	}

    }
 moveto(3, 193);		/* Line 3-20 */
 outtext("");
 moveto(3, 203);		/* Line 3-21 */
 outtext("Source Program available; No-cost license, boblark@proaxis.com");
 moveto(3, 213);		/* Line 3-22 */
 outtext("Program Copyright (c) Robert Larkin 1996, 1997, 1998, 1999");


	for(;;)			/* Wait for keyboard hit */
	    if ( kbhit() )
		{_bios_keybrd(_KEYBRD_READ);  break; }
	setviewport(X2L+1, Y2T+96, X2R-1, Y2B-12, DO_CLIP);
	clearviewport();
	putimage(0,   0, big_area_ptr1, COPY_PUT);
	putimage(0, 161, big_area_ptr2, COPY_PUT);
	}

/* left_justify(char *str1, char *str2, int len2, char fill)
 * str1 is copied to str2 and the space to the right is filled with spaces.
 * Therefore str1 should not be longer than len2 or it will be truncated.
 * The difference between str1 and str2 is right-filled with 'fill'.
 * str2 is terminated with a null and must be declared as len2+1 or more.
 */
char *left_justify(char *str1, char *str2, int len2, char fill)
    {
    int len1, l, i;

    len1=strlen(str1);
    if (len1>len2)			/* Won't fit case!  */
	l=len2;
    else
	{
	l=len1;
	for (i=len1; i<len2; i++)	/* Fill to end with blanks */
	    *(str2+i)=fill;
	}
    for (i=0; i<l; i++)			/* Move the string into place */
	*(str2+i)=*(str1+i);
    if (l < len2)
	*(str2+l)=32;			/* Single blank at end; readability */
    *(str2+len2)=0;			/* Terminate the string in a null */
    return str2;
    }

/* Alternate error handlers  so that program does not exit to dos from
   Ctrl-C or from floating point error */
void  ctrlc_handler(int sig)
	{
	int i;
	do_error(200, 100);		/* Error code 200 */
	signal(SIGINT, ctrlc_handler);  /* Reinstall for next time */
	i=sig;  /* Stop warning from compiler */
	}

void  fpe_handler(int sig)
	{
	int i;
	do_error(201, 100);
	signal(SIGFPE, fpe_handler);
	i=sig;
	}

void display_summary(char *fname, char *ver)
	{
	char cc[100], *pc;

	strcpy(cc, fname);		/* Get a copy to play with */
	cc[strlen(cc)-4]=0;		/* Remove the .exe on the end */
	pc=&cc[strlen(cc)];		/* Now look for end of path */
	while (*pc!='\\' && pc!=&cc[0])  pc--;
	pc++;				/* Move past the slash */
	printf("\nUHFA Transceiver Controller v%s  de W7PUA\n\n", ver);
	printf("Your command line lacked necessary information.\n");
	printf("Use '%s com' where com is a number between 1 and\n", pc);
	printf("4 that sets the COM port to be used.\n");
	printf("Thus for COM2 enter\n");
	printf("        %s 2 \n\n", pc);
	}

void license_info(void)
    {
    clrscr();
    print_stars();
    printf("\n\n\t     2-Meter Transceiver DSP-10 (uhfa.exe) v1.60\n\n");
    printf("\t\t\t     12 Sept 1999\n");
    printf("\n\t  Copyright (c) 1996, 1997, 1998, 1999 Robert S. Larkin\n\n");
    print_stars();  printf("\n\n");

printf("\nDSP-10 (uhfa.exe) comes with ABSOLUTELY NO WARRANTY; for details type `1'.");
printf("\nThis is free software, and you are welcome to redistribute it");
printf("\nunder certain conditions; type `2' for details, or alternatively, read");
printf("\nthe plain text file GNU_GPL.txt that is in the same directory with");
printf("\nthis executable file uhfa.exe. Be sure to read it.\n");
printf("\nHit the 'Enter' key (Return key) to continue with the DSP-10 program.");
    }

void warranty(void)
    {
    clrscr();
printf("\n\n                 NO WARRANTY\n");
printf("\nBECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY");
printf("\nFOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN");
printf("\nOTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES");
printf("\nPROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED");
printf("\nOR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF");
printf("\nMERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS");
printf("\nTO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE");
printf("\nPROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,");
printf("\nREPAIR OR CORRECTION.\n");
printf("\nIN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING");
printf("\nWILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR");
printf("\nREDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,");
printf("\nINCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING");
printf("\nOUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED");
printf("\nTO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY");
printf("\nYOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER");
printf("\nPROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE");
printf("\nPOSSIBILITY OF SUCH DAMAGES.\n");
printf("\nAfter reading, hit Enter (Return) to continue");
    getch();
    }

void license(void)
    {
    char cline[90];
    int nl;

    nl=0;
    clrscr();
    if (NULL == ( license_file=fopen("GNU_GPL.txt", "r") ))
	    {
	    printf("\nNo GNU_GPL.txt license file was found. This must be in");
	    printf(" the same directory\nas UHFA.EXE file. Exiting program.");
	    exit(1);
	    }
    while (fgets(cline, 80, license_file))
	{
	printf("%s", cline);
	if (nl++ >= 20)
	    {
	    nl=0;
	    printf("\n\n(After reading, hit any key to continue)");
	    getch();
	    clrscr();
	    }
	}
    fclose(license_file);
    printf("\n\n(After reading, hit any key to continue)");
    getch();
    clrscr();
    }


void quit(void)
	{               
                fclose(sun_file);
                fclose(cold_file);
	free( big_area_ptr1 );
	free( big_area_ptr2 );
	if (transmit)  t_r_toggle();  /* Leave hardware in rcv */
	if (quit_save_state)
	    save_state();
	ASYNCSTOP();
	setallpalette(&savpalette);	/* The palette when we got here */
	setbkcolor(BLACK);
	setcolor(UWHITE);
	closegraph();
	printf("Exiting through quit()\n");
	exit(0);
	}
